"""LBFGS for solving non-linear minimization problems."""

from collections.abc import Callable, Sequence
from typing import Literal

import torch
from torch.optim import LBFGS
from typing_extensions import Unpack

from mrpro.algorithms.optimizers.OptimizerStatus import OptimizerStatus
from mrpro.operators.Operator import OperatorType


class LBFGSStatus(OptimizerStatus):
    """Status of the LBFGS algorithm."""

    residual: torch.Tensor
    """Residual of the current estimate."""


def lbfgs(
    f: OperatorType | Callable[[Unpack[tuple[torch.Tensor, ...]]], tuple[torch.Tensor]],
    initial_parameters: Sequence[torch.Tensor],
    learning_rate: float = 1.0,
    max_iterations: int = 100,
    max_evaluations: int | None = 100,
    tolerance_grad: float = 1e-07,
    tolerance_change: float = 1e-09,
    history_size: int = 10,
    line_search_fn: None | Literal['strong_wolfe'] = 'strong_wolfe',
    callback: Callable[[OptimizerStatus], bool | None] | None = None,
) -> tuple[torch.Tensor, ...]:
    r"""
    LBFGS for (non-linear) minimization problems.

    The Limited-memory Broyden-Fletcher-Goldfarb-Shanno (LBFGS) algorithm is a quasi-Newton optimization method
    that approximates the inverse Hessian matrix using a limited memory of past gradients and updates.
    It is well-suited for high-dimensional problems and leverages curvature information
    for faster convergence compared to first-order methods such as `mrpro.algorithms.optimizers.adam`.

    The parameter update rule is:

    .. math::

        \theta_{k+1} = \theta_k - \alpha_k H_k \nabla f(\theta_k),

    where :math:`H_k` is a limited-memory approximation of the inverse Hessian,
    and :math:`\alpha_k` is the step size determined via line search (e.g., strong Wolfe conditions).

    The algorithm performs the following steps:

    1. Compute the gradient of the objective function.
    2. Approximate the inverse Hessian matrix :math:`H_k` using stored gradients and updates.
    3. Perform a line search to compute the step size :math:`\alpha_k`.
    4. Update the parameters.
    5. Store the latest gradient and update information.

    This implementation wraps PyTorch's `torch.optim.LBFGS` class.
    For more information, see [WIKI]_, [NOC1980]_, and [LIU1989]_.

    References
    ----------
    .. [NOC1980] Nocedal, J. (1980). "Updating quasi-Newton matrices with limited storage."
       *Mathematics of Computation*, 35(151), 773-782. https://doi.org/10.1090/S0025-5718-1980-0572855-7
    .. [LIU1989] Liu, D. C., & Nocedal, J. (1989). "On the limited memory BFGS method for large scale optimization."
       *Mathematical Programming*, 45(1-3), 503-528. https://doi.org/10.1007/BF01589116
    .. [WIKI] Wikipedia: Limited-memory_BFGS https://en.wikipedia.org/wiki/Limited-memory_BFGS


    Parameters
    ----------
    f
        scalar function to be minimized
    initial_parameters
        `Sequence` of parameters to be optimized.
        Note that these parameters will not be changed. Instead, we create a copy and
        leave the initial values untouched.
    learning_rate
        learning rate. This should usually be left as ``1.0`` if a line search is used.
    max_iterations
        maximal number of iterations
    max_evaluations
        maximal number of evaluations of `f` per optimization step
    tolerance_grad
        termination tolerance on first order optimality
    tolerance_change
        termination tolerance on function value/parameter changes
    history_size
        update history size
    line_search_fn
        line search algorithm, either ``strong_wolfe`` or `None` (meaning constant step size)
    callback
        Function to be called after each iteration. This can be used to monitor the progress of the algorithm.
        If it returns `False`, the algorithm stops at that iteration.
        N.B. the callback is not called within the line search of LBFGS.
        You can use the information from the `~mrpro.algorithms.optimizers.OptimizerStatus`
        to display a progress bar.

    Returns
    -------
        List of optimized parameters.
    """
    parameters = tuple(p.detach().clone().requires_grad_(True) for p in initial_parameters)
    optim = LBFGS(
        params=parameters,
        lr=learning_rate,
        history_size=history_size,
        max_iter=max_iterations,
        max_eval=max_evaluations,
        tolerance_grad=tolerance_grad,
        tolerance_change=tolerance_change,
        line_search_fn=line_search_fn,
    )

    iteration = 0

    def closure():
        nonlocal iteration
        optim.zero_grad()
        (residual,) = f(*parameters)
        residual.backward()

        if callback is not None:
            state = optim.state[optim.param_groups[0]['params'][0]]
            if state['n_iter'] > iteration:
                # true in a new iteration, false in line search
                status = LBFGSStatus(
                    solution=tuple(p.detach() for p in parameters),
                    residual=residual.detach(),
                    iteration_number=state['n_iter'],  # status is from previous iteration
                )
                if callback(status) is False:
                    raise StopIteration
                iteration += 1

        return residual

    try:  # noqa: SIM105
        # This does NOT only perform a single step, but runs the full optimization.
        # See https://pytorch.org/docs/stable/generated/torch.optim.LBFGS.html
        optim.step(closure)
    except StopIteration:
        # callback asked us to stop.
        ...

    parameters = tuple(p.detach() for p in parameters)
    return parameters



"""Zero pad and crop data tensor."""

import math
from collections.abc import Sequence
from typing import Literal

import torch

from mrpro.utils.indexing import normalize_index
from mrpro.utils.reshape import unsqueeze_left


def pad_or_crop(
    data: torch.Tensor,
    new_shape: Sequence[int] | torch.Size,
    dim: None | Sequence[int] = None,
    mode: Literal['constant', 'reflect', 'replicate', 'circular'] = 'constant',
    value: float = 0.0,
) -> torch.Tensor:
    """Change shape of data by center cropping or symmetric padding.

    Parameters
    ----------
    data
        Data to pad or crop.
    new_shape
        Desired shape of data.
    dim
        Dimensions the `new_shape` corresponds to.
        `None` is interpreted as last ``len(new_shape)`` dimensions.
    mode
        Mode to use for padding.
    value
        Value to use for constant padding.

    Returns
    -------
        Data zero padded or cropped to shape.
    """
    if len(new_shape) > data.ndim:
        raise ValueError('length of new shape should not exceed dimensions of data')

    if dim is None:  # Use last dimensions
        new_shape = (*data.shape[: -len(new_shape)], *new_shape)
    else:
        if len(new_shape) != len(dim):
            raise ValueError('length of shape should match length of dim')
        dim = tuple(normalize_index(data.ndim, idx) for idx in dim)  # raises if any not in [-data.ndim,data.ndim)
        if len(dim) != len(set(dim)):  # this is why we normalize
            raise ValueError('repeated values are not allowed in dims')
        # Update elements in data.shape at indices specified in dim with corresponding elements from new_shape
        new_shape = tuple(new_shape[dim.index(i)] if i in dim else s for i, s in enumerate(data.shape))

    npad: list[int] = []
    for old, new in zip(data.shape, new_shape, strict=True):
        diff = new - old
        after = math.trunc(diff / 2)
        before = diff - after
        if before or after or npad:
            npad.append(before)
            npad.append(after)

    n_extended_dims = 0
    if mode != 'constant':
        # See https://docs.pytorch.org/docs/stable/generated/torch.nn.functional.pad.html for supported shapes.
        while len(npad) // 2 < data.ndim - 2:
            npad = [0, 0, *npad]

        n_extended_dims = max(0, len(npad) // 2 - (data.ndim - 2))
        if n_extended_dims:  # We need to extend data such that the padding is supported.
            data = unsqueeze_left(data, n_extended_dims)

        if len(npad) > 6:  # TODO: reshape and call multiple times
            raise ValueError('Non-constant padding is only supported for up to the last 3 dimensions.')

    if any(npad):
        # F.pad expects paddings in reversed order
        data = torch.nn.functional.pad(data, npad[::-1], value=value, mode=mode)

    if n_extended_dims:
        idx = n_extended_dims * (0,)
        data = data[idx]
    return data



"""Regularized Iterative SENSE Reconstruction."""

from __future__ import annotations

from collections.abc import Callable

import torch

from mrpro.algorithms.optimizers.cg import cg
from mrpro.algorithms.prewhiten_kspace import prewhiten_kspace
from mrpro.algorithms.reconstruction.DirectReconstruction import DirectReconstruction
from mrpro.data.CsmData import CsmData
from mrpro.data.DcfData import DcfData
from mrpro.data.IData import IData
from mrpro.data.KData import KData
from mrpro.data.KNoise import KNoise
from mrpro.operators.IdentityOp import IdentityOp
from mrpro.operators.LinearOperator import LinearOperator
from mrpro.utils import unsqueeze_right


class RegularizedIterativeSENSEReconstruction(DirectReconstruction):
    r"""Regularized iterative SENSE reconstruction.

    This algorithm solves the problem :math:`\min_{x} ||Ax - y||_2^2 + \lambda||Bx - x_{reg}||_2^2`
    by using a conjugate gradient algorithm to solve
    :math:`H x = b` with :math:`H = A^H A + \lambda B^H B` and :math:`b = A^H y + \lambda B^H x_{reg}` where :math:`A`
    is the acquisition model (coil sensitivity maps, Fourier operator, k-space sampling), :math:`y` is the acquired
    k-space data, :math:`\lambda` is the strength of the regularization, and :math:`x_{reg}` is the regularization image
    (i.e. a prior). :math:`B` is a linear operator applied to :math:`x`.
    """

    n_iterations: int
    """Number of CG iterations."""

    regularization_data: torch.Tensor
    r"""Regularization data (i.e. prior) :math:`x_{\mathrm{reg}}`."""

    regularization_weight: torch.Tensor
    r"""Strength of the regularization :math:`\lambda`."""

    regularization_op: LinearOperator
    """Linear operator :math:`B` applied to the current estimate in the regularization term."""

    def __init__(
        self,
        kdata: KData | None = None,
        fourier_op: LinearOperator | None = None,
        csm: Callable | CsmData | None = CsmData.from_idata_walsh,
        noise: KNoise | None = None,
        dcf: DcfData | None = None,
        *,
        n_iterations: int = 5,
        regularization_data: float | torch.Tensor = 0.0,
        regularization_weight: float | torch.Tensor,
        regularization_op: LinearOperator | None = None,
    ) -> None:
        r"""Initialize RegularizedIterativeSENSEReconstruction.

        For a unregularized version of the iterative SENSE algorithm the regularization_weight can be set to ``0``
        or `~mrpro.algorithms.reconstruction.IterativeSENSEReconstruction` algorithm can be used.

        Parameters
        ----------
        kdata
            If `kdata` is provided and `fourier_op` or `dcf` are `None`, then `fourier_op` and `dcf` are estimated
            based on `kdata`. Otherwise `fourier_op` and `dcf` are used as provided.
        fourier_op
            Instance of the `~mrpro.operators.FourierOp` used for reconstruction.
            If `None`, set up based on `kdata`.
        csm
            Sensitivity maps for coil combination. If `None`, no coil combination is carried out, i.e. images for each
            coil are returned. If a `Callable` is provided, coil images are reconstructed using the adjoint of the
            `~mrpro.operators.FourierOp` (including density compensation) and then sensitivity maps are calculated
            using the `Callable`. For this, `kdata` needs also to be provided.
            For examples have a look at the `mrpro.data.CsmData` class e.g. `~mrpro.data.CsmData.from_idata_walsh`
            or `~mrpro.data.CsmData.from_idata_inati`.
        noise
            Noise used for prewhitening. If `None`, no prewhitening is performed
        dcf
            K-space sampling density compensation. If `None`, set up based on `kdata`.
            Used to obtain a the starting point for the CG algorithm as the scaled density compensated direct
            reconstruction [FESSLER2010]_.
        n_iterations
            Number of CG iterations
        regularization_data
            Regularization data, e.g. a reference image (:math:`x_0`).
        regularization_weight
            Strength of the regularization (:math:`\lambda`).
        regularization_op
            Linear operator :math:`B` applied to the current estimate in the regularization term. If None, nothing is
            applied to the current estimate.

        References
        ----------
        .. [FESSLER2010] Fessler, J.A., Noll, D.C: Iterative Reconstruction Methods for Non-Cartesian MRI.
           https://ece-classes.usc.edu/ee591/library/Fessler-Iterative%20Reconstruction.pdf

        Raises
        ------
        `ValueError`
            If the `kdata` and `fourier_op` are `None` or if `csm` is a `Callable` but `kdata` is None.
        """
        super().__init__(kdata, fourier_op, csm, noise, dcf)
        self.n_iterations = n_iterations
        self.regularization_data = torch.as_tensor(regularization_data)
        self.regularization_weight = torch.as_tensor(regularization_weight)
        self.regularization_op = regularization_op if regularization_op is not None else IdentityOp()

    def forward(self, kdata: KData) -> IData:
        """Apply the reconstruction.

        Parameters
        ----------
        kdata
            k-space data to reconstruct.

        Returns
        -------
            the reconstruced image.
        """
        if self.noise is not None:
            kdata = prewhiten_kspace(kdata, self.noise)

        acquisition_model = self.fourier_op
        if self.csm is not None:
            acquisition_model = acquisition_model @ self.csm.as_operator()

        operator = acquisition_model.gram
        (right_hand_side,) = acquisition_model.H(kdata.data)

        if not torch.all(self.regularization_weight == 0):  # Has Regularization
            operator = operator + self.regularization_weight * self.regularization_op.gram
            right_hand_side = (
                right_hand_side + self.regularization_weight * self.regularization_op.H(self.regularization_data)[0]
            )

        if self.dcf is not None:
            # The DCF is used to obtain a good starting point for the CG algorithm.
            # This is equivalten to running the CG algorithm with H = A^H DCF A and b = A^H DCF y
            # for a single iteration.
            dcf_op = self.dcf.as_operator()
            (u,) = (acquisition_model.H @ dcf_op)(kdata.data)
            (v,) = (acquisition_model.H @ dcf_op @ acquisition_model)(u)
            u_flat = u.flatten(start_dim=-3)
            v_flat = v.flatten(start_dim=-3)
            initial_value = (
                unsqueeze_right(torch.linalg.vecdot(u_flat, u_flat) / torch.linalg.vecdot(v_flat, u_flat), 3) * u
            )
        else:
            # The right and side is not a good starting point without DCF.
            initial_value = torch.zeros_like(right_hand_side)
        (img_tensor,) = cg(
            operator,
            right_hand_side,
            initial_value=initial_value,
            max_iterations=self.n_iterations,
            tolerance=0,  # run for a fixed number of iterations
        )
        img = IData.from_tensor_and_kheader(img_tensor, kdata.header)
        return img



"""Brainweb Phantom."""

import concurrent.futures
import gzip
import io
import re
import time
from collections.abc import Callable, Mapping, Sequence
from dataclasses import dataclass
from os import PathLike
from pathlib import Path
from types import MappingProxyType
from typing import Literal

import h5py
import numpy as np
import platformdirs
import requests
import torch
import torchvision.transforms.functional
from tqdm import tqdm
from typing_extensions import TypeVar

OVERVIEW_URL = 'http://brainweb.bic.mni.mcgill.ca/brainweb/anatomic_normal_20.html'
URL_TEMPLATE = (
    'http://brainweb.bic.mni.mcgill.ca/cgi/brainweb1?'
    'do_download_alias=subject{subject}_{c}'
    '&format_value=raw_short'
    '&zip_value=gnuzip'
    '&download_for_real=%5BStart+download%21%5D'
)

# includes background
ALL_CLASSES = ('bck', 'skl', 'gry', 'wht', 'csf', 'mrw', 'dura', 'fat', 'fat2', 'mus', 'm-s', 'ves')  # noqa: typos
VERSION = 1
# ~/.cache/mrpro/brainweb on Linux, %AppData%\Local\mrpro\brainweb on Windows
CACHE_DIR_BRAINWEB = Path(platformdirs.user_cache_dir('mrpro')) / 'brainweb'
K = TypeVar('K')
TClassNames = Literal['skl', 'gry', 'wht', 'csf', 'mrw', 'dura', 'fat', 'fat2', 'mus', 'm-s', 'ves']  # noqa: typos
BRAINWEBSHAPE = (362, 434, 362)


@dataclass
class BrainwebTissue:
    """Container for parameters of a single tissue.

    Attributes are either single values or ranges.
    If ranges, values are sampled uniformly from within this range by the `sample_r1`, `sample_r2`, and `sample_m0`
    methods. If a single value is given, this value is returned by the respective method.
    """

    t1: float | tuple[float, float]
    """T1 value or range (T1_min, T1_max) in seconds."""

    t2: float | tuple[float, float]
    """T2 value or range (T2_min, T2_max) in seconds."""

    m0_abs: float | tuple[float, float]
    """Absolute value or range (M0_min, M0_max) of the complex M0."""

    m0_phase: float | tuple[float, float] = 0.0
    """Phase value or range (Phase_min, Phase_max) of the complex M0 in radians."""

    def sample_r1(self, rng: None | torch.Generator = None) -> torch.Tensor:
        """Get (possibly randomized) r1=1/t1 value.

        Parameters
        ----------
        rng
            Random number generator. `None` uses the default generator.
        """
        if isinstance(self.t1, tuple):
            return 1 / torch.empty(1).uniform_(*self.t1, generator=rng)
        return 1 / torch.tensor(self.t1)

    def sample_r2(self, rng: None | torch.Generator = None) -> torch.Tensor:
        """Get (possibly randomized) r2=1/t2 value.

        Parameters
        ----------
        rng
            Random number generator. `None` uses the default generator.
        """
        if isinstance(self.t2, tuple):
            return 1 / torch.empty(1).uniform_(*self.t2, generator=rng)
        return 1 / torch.tensor(self.t2)

    def sample_m0(self, rng: None | torch.Generator = None) -> torch.Tensor:
        """Get (possibly randomized) complex m0 value.

        Parameters
        ----------
        rng
            Random number generator. `None` uses the default generator.
        """
        if isinstance(self.m0_abs, tuple):
            magnitude = torch.empty(1).uniform_(*self.m0_abs, generator=rng)
        else:
            magnitude = torch.tensor(self.m0_abs)
        if isinstance(self.m0_phase, tuple):
            phase = torch.empty(1).uniform_(*self.m0_phase, generator=rng)
        else:
            phase = torch.tensor(self.m0_phase)
        return torch.polar(magnitude, phase)


def augment(
    size: int = 256,
    trim: bool = True,
    max_random_shear: float = 5,
    max_random_rotation: float = 10,
    max_random_scaling_factor: float = 0.1,
    p_horizontal_flip: float = 0.5,
    p_vertical_flip: float = 0.5,
) -> Callable[[torch.Tensor, torch.Generator | None], torch.Tensor]:
    """Get an augmentation function.

    Creates a function that applies augmentation to the input tensor consisting of
    rotation, shearing, scaling and horizontal/vertical flipping.

    The image is scaled such that the largest dimension is in
    [size * (1 - max_random_scaling), size * (1 + max_random_scaling)], then padded/cropped to size `size x size`.
    In scaling, the aspect ratio is preserved.
    Random horizontal and vertical flips are applied with probability `p_horizontal_flip` and `p_vertical_flip`.

    Parameters
    ----------
    size
        resulting image will be (size x size) pixels.
    trim
        If True, remove fully zero outer rows and columns before scaling
    max_random_shear
        Maximum random shear in degrees, shear is in [-max_shear, max_shear] in x and y direction.
    max_random_rotation
        Maximum random rotation in degrees, rotation is in [-max_rotation, max_rotation].
    max_random_scaling_factor
        Strength of the scaling randomization (see above).
    p_horizontal_flip
        Probability of horizontal flip.
    p_vertical_flip
        Probability of vertical flip.

    Returns
    -------
        Callable that applies augmentation to the input tensor
    """

    def augment_fn(data: torch.Tensor, rng: torch.Generator | None = None) -> torch.Tensor:
        """Apply augmentation to the input tensor."""
        rand = torch.empty(6).uniform_(-1, 1, generator=rng).tolist()

        shear_x = rand[0] * max_random_shear
        shear_y = rand[1] * max_random_shear
        angle = rand[2] * max_random_rotation
        scale = size / max(data.shape[-2:])
        scale *= 1 + max_random_scaling_factor * rand[3]
        translate = rand[4:6]  # subpixel translation for edge aliasing
        if trim:
            data = data[(slice(None), *trim_indices(data.sum(0) > 0.1 * data.amax()))]

        data = torchvision.transforms.functional.affine(
            data,
            angle=angle,
            scale=scale,
            shear=[shear_x, shear_y],
            translate=translate,
            interpolation=torchvision.transforms.InterpolationMode.BILINEAR,
            fill=0.0,
        )
        rand = torch.empty(2).uniform_(0, 1, generator=rng).tolist()
        data = torchvision.transforms.functional.hflip(data) if rand[0] < p_horizontal_flip else data
        data = torchvision.transforms.functional.vflip(data) if rand[1] < p_vertical_flip else data

        data = torchvision.transforms.functional.center_crop(data, [size, size])

        return data

    return augment_fn


def resize(size: int = 256) -> Callable[[torch.Tensor, torch.Generator | None], torch.Tensor]:
    """Get a resizing and cropping function.

    Parameters
    ----------
    size
        Size of the output tensor.

    Returns
    -------
        Callable that resizes and crops the input tensor.
    """

    def resize_fn(data: torch.Tensor, rng: torch.Generator | None = None) -> torch.Tensor:  # noqa: ARG001
        """Apply resizing and cropping to the input tensor."""
        scale = size / max(data.shape[-2:])
        data = torchvision.transforms.functional.resize(data, [int(scale * data.shape[1]), int(scale * data.shape[2])])
        data = torchvision.transforms.functional.center_crop(data, [size, size])
        return data

    return resize_fn


DEFAULT_AUGMENT_256 = augment(256)
"""Default random augmentation for 256x256 images."""


def trim_indices(mask: torch.Tensor) -> tuple[slice, slice]:
    """Get slices that remove fully masked out outer rows and columns.

    Parameters
    ----------
    mask
        Mask indicating valid data.

    Returns
    -------
        Two `slice` objects, that can be used to index the data
        to remove fully masked out outer rows and columns.
    """
    mask = mask.any(dim=tuple(range(mask.ndim - 2)))
    row_mask, col_mask = mask.any(1).short(), mask.any(0).short()
    row_min = int(torch.argmax(row_mask))
    row_max = int(mask.size(0) - torch.argmax(row_mask.flip(0)))
    col_min = int(torch.argmax(col_mask))
    col_max = int(mask.size(1) - torch.argmax(col_mask.flip(0)))
    return slice(row_min, row_max), slice(col_min, col_max)


VALUES_3T_RANDOMIZED: Mapping[TClassNames, BrainwebTissue] = MappingProxyType(
    {
        'skl': BrainwebTissue((0.000, 2.000), (0.000, 0.010), (0.00, 0.05), (-0.2, 0.2)),
        'gry': BrainwebTissue((1.200, 2.000), (0.080, 0.120), (0.70, 1.00), (-0.2, 0.2)),
        'wht': BrainwebTissue((0.800, 1.500), (0.060, 0.100), (0.50, 0.90), (-0.2, 0.2)),  # noqa:typos
        'csf': BrainwebTissue((2.000, 4.000), (1.300, 2.000), (0.90, 1.00), (-0.2, 0.2)),
        'mrw': BrainwebTissue((0.400, 0.600), (0.060, 0.100), (0.70, 1.00), (-0.2, 0.2)),
        'dura': BrainwebTissue((2.000, 2.800), (0.200, 0.500), (0.90, 1.00), (-0.2, 0.2)),
        'fat': BrainwebTissue((0.300, 0.500), (0.060, 0.100), (0.90, 1.00), (-0.2, 0.2)),
        'fat2': BrainwebTissue((0.400, 0.600), (0.060, 0.100), (0.60, 0.90), (-0.2, 0.2)),
        'mus': BrainwebTissue((1.200, 1.500), (0.040, 0.060), (0.90, 1.00), (-0.2, 0.2)),
        'm-s': BrainwebTissue((0.500, 0.900), (0.300, 0.500), (0.90, 1.00), (-0.2, 0.2)),
        'ves': BrainwebTissue((1.700, 2.100), (0.200, 0.400), (0.80, 1.00), (-0.2, 0.2)),
    }
)
"""Tissue values for 3T with wide randomization ranges."""

DEFAULT_VALUES = {'r1': 0.0, 'm0': 0.0, 'r2': 0.0, 'mask': 0, 'tissueclass': -1, 't1': 10.0, 't2': 0.0}
"""Default values for masked out regions."""


def download_brainweb(
    output_directory: str | PathLike = CACHE_DIR_BRAINWEB,
    workers: int = 4,
    progress: bool = False,
    compress: bool = False,
    n_files: int = 20,
) -> None:
    """Download Brainweb data.

    Parameters
    ----------
    output_directory
        Directory to save the data.
    workers
        Number of parallel downloads.
    progress
        Show progress bar.
    compress
        Use compression for HDF5 files. Saves disk space but might slow down (or speed up) access,
        depending on the system and access pattern.
    n_files
        Number of files to download. Can be used to download only a subset of the files.
        Cannot be greater than 20.
    """
    if n_files > 20:
        raise ValueError('Cannot download more than 20 files.')

    def load_file(
        url: str,
        timeout: float = 60,
        max_retries: int = 3,
        retry_delay: float = 30,
    ) -> bytes:
        """Load url content with retries for network errors."""
        for attempt in range(max_retries):
            try:
                response = requests.get(url, timeout=timeout)
                response.raise_for_status()
                return response.content
            except requests.exceptions.RequestException:
                if attempt == max_retries - 1:
                    break
                time.sleep(retry_delay)
        raise ConnectionError(f'Failed to download {url} after {max_retries} attempts.')

    def unpack(data: bytes, dtype: np.typing.DTypeLike, shape: Sequence[int]) -> np.ndarray:
        """Unpack gzipped data."""
        return np.frombuffer(gzip.open(io.BytesIO(data)).read(), dtype=dtype).reshape(shape)

    def norm_(values: list[np.ndarray]) -> np.ndarray:
        """Normalize values to sum to 1 and convert to uint8."""
        for i, x in enumerate(values):
            values[i] = np.clip(x - np.min(x[50], (0, 1)), 0, 4096)
        sum_values = sum(values)
        values.pop(ALL_CLASSES.index('bck'))  # noqa: typos
        for i, x in enumerate(values):
            x = np.divide(x, sum_values, out=np.zeros_like(x, dtype=float), where=sum_values != 0)
            x[sum_values == 0] = 0
            x = (x * (2**8 - 1)).astype(np.uint8)
            values[i] = x
        return np.stack(values, -1)

    def download_subject(subject: str, outfilename: Path, workers: int, progressbar: tqdm) -> None:
        """Download and process all class files for a single subject asynchronously."""
        with concurrent.futures.ThreadPoolExecutor(max_workers=workers) as executor:
            futures = {executor.submit(load_file, URL_TEMPLATE.format(subject=subject, c=c)): c for c in ALL_CLASSES}

            downloaded_data = {}
            for future in concurrent.futures.as_completed(futures):
                c = futures[future]
                downloaded_data[c] = future.result()
                progressbar.update(1)

        values = norm_([unpack(downloaded_data.pop(c), shape=BRAINWEBSHAPE, dtype=np.uint16) for c in ALL_CLASSES])

        with h5py.File(outfilename, 'w') as f:
            f.create_dataset(
                'classes',
                values.shape,
                dtype=values.dtype,
                data=values,
                chunks=(4, 4, 4, values.shape[-1]) if compress else None,
                compression='lzf' if compress else None,
            )
            f.attrs['classnames'] = [c for c in ALL_CLASSES if c != 'bck']  # noqa: typos
            f.attrs['subject'] = int(subject)
            f.attrs['version'] = VERSION

    page = requests.get(OVERVIEW_URL, timeout=5)
    subjects = re.findall(r'option value=(\d*)>', page.text)[:n_files]
    output_directory = Path(output_directory)
    output_directory.mkdir(parents=True, exist_ok=True)

    totalsteps = len(subjects) * len(ALL_CLASSES)
    with tqdm(total=totalsteps, desc='Downloading Brainweb data', disable=not progress) as progressbar:
        for subject in subjects:
            outfilename = output_directory / f's{subject}.h5'
            if outfilename.exists():
                with h5py.File(outfilename, 'r') as f:
                    if (
                        f.attrs.get('version', 0) == VERSION
                        and f.attrs.get('subject', 0) == int(subject)
                        and f['classes'].compression == 'lzf'
                        if compress
                        else None
                    ):
                        # file is already downloaded and up to date
                        progressbar.update(len(ALL_CLASSES))
                        continue
            download_subject(subject, outfilename, workers, progressbar)


class BrainwebVolumes(torch.utils.data.Dataset):
    """Dataset of 3D qMRI parameters based on Brainweb dataset.

    This dataset provides 1 mm isotropic 3D brain data of various quantitative MRI (qMRI)
    parameters based on the segmentations provided by the Brainweb [AubertBroche2006]_ dataset.

    References
    ----------
    .. [AubertBroche2006] Aubert-Broche, B., Griffin, M., Pike, G.B., Evans, A.C., & Collins, D.L. (2006).
       Twenty New Digital Brain Phantoms for Creation of Validation Image Data Bases.
       IEEE Transactions on Medical Imaging, 25 (11), 1410-1416. https://doi.org/10.1109/TMI.2006.883453
    """

    @staticmethod
    def download(
        output_directory: str | PathLike = CACHE_DIR_BRAINWEB,
        workers: int = 4,
        progress: bool = False,
        compress: bool = False,
    ) -> None:
        """Download Brainweb data.

        Parameters
        ----------
        output_directory
            Directory to save the data.
        workers
            Number of parallel downloads.
        progress
            Show progress bar.
        compress
            Use compression for HDF5 files. Saves disk space but might slow down or speed up access,
            depending on the file system and access pattern.

        """
        download_brainweb(output_directory, workers, progress, compress)

    def __init__(
        self,
        folder: str | Path = CACHE_DIR_BRAINWEB,
        what: Sequence[Literal['r1', 'r2', 'm0', 't1', 't2', 'mask', 'tissueclass'] | TClassNames] = ('m0', 'r1', 'r2'),
        parameters: Mapping[TClassNames, BrainwebTissue] = VALUES_3T_RANDOMIZED,
        mask_values: Mapping[str, float | None] = DEFAULT_VALUES,
        seed: int | Literal['index', 'random'] = 'random',
    ) -> None:
        """Initialize Dataset.

        Parameters
        ----------
        folder:
            The directory containing Brainweb HDF5 files
        what
            What to return for each subject.
            Possible values are:
                - r1: R1 relaxation rate.
                - r2: R2 relaxation rate.
                - m0: M0 magnetization.
                - t1: T1 relaxation time.
                - t2: T2 relaxation time.
                - mask: mask indicating valid data.
                - tissueclass: (majority) class index.
                - Brainweb class name: raw percentage of the specific tissue class (see below for possible values).
        parameters
            Parameters for each tissue class.
            The Brainweb tissue classes are:
                - skl: skull
                - gry: gray matter
                - wht: white matter
                - csf: cerebrospinal fluid
                - mrw: bone marrow
                - dura: dura
                - fat: fat
                - fat2: fat and tissue
                - mus: muscle
                - m-s: skin
                - ves: vessels
        mask_values
            Default values to use for masked out regions.
        seed
            Random seed. Can be an int, the strings ``index`` to use subject index as seed,
            or ``random`` for random seed.
        """
        self.files = list(Path(folder).glob('s??.h5'))

        if not self.files:
            raise FileNotFoundError(f'No files found in {folder}.')
        self.parameters = parameters
        self.what = what

        if seed == 'random':
            self._rng: torch.Generator | None = torch.default_generator
        elif isinstance(seed, int):
            self._rng = torch.Generator().manual_seed(seed)
        elif seed == 'index':
            self._rng = None

        self.mask_values = mask_values

    def __len__(self) -> int:
        """Get number of subjects."""
        return len(self.files)

    def __getitem__(
        self, index: int
    ) -> dict[Literal['r1', 'r2', 'm0', 't1', 't2', 'mask', 'tissueclass'] | TClassNames, torch.Tensor]:
        """Get 3D data from one subject."""
        with h5py.File(self.files[index]) as file:
            data = torch.as_tensor(np.array(file['classes'], dtype=np.uint8))
            classnames = tuple(file.attrs['classnames'])
        result: dict[Literal['r1', 'r2', 'm0', 't1', 't2', 'mask', 'tissueclass'] | TClassNames, torch.Tensor] = {}
        for el in self.what:
            if el == 'r1':
                # / 255 to convert from uint8 to 0...1
                values = torch.stack([self.parameters[k].sample_r1() for k in classnames]) / 255
                result[el] = (data.to(values) @ values)[..., 0]
            elif el == 'r2':
                values = torch.stack([self.parameters[k].sample_r2() for k in classnames]) / 255
                result[el] = (data.to(values) @ values)[..., 0]
            elif el == 'm0':
                values = torch.stack([self.parameters[k].sample_m0() for k in classnames]) / 255
                result[el] = (data.to(values) @ values)[..., 0]
            elif el == 't1':
                values = torch.stack([self.parameters[k].sample_r1() for k in classnames]) / 255
                result[el] = (data.to(values) @ values)[..., 0].reciprocal()
            elif el == 't2':
                values = torch.stack([self.parameters[k].sample_r2() for k in classnames]) / 255
                result[el] = (data.to(values) @ values)[..., 0].reciprocal()
            elif el == 'tissueclass':
                result[el] = data.argmax(-1)
            elif el in classnames:
                result[el] = data[..., classnames.index(el)] / 255
            elif el == 'mask':
                mask = (data < 20).all(-1)
                mask = (
                    torch.nn.functional.conv3d(mask[None, None].float(), torch.ones(1, 1, 3, 3, 3), padding=1)[0, 0] < 1
                )
                result[el] = ~mask
            else:
                raise NotImplementedError(f'what=({el},) is not implemented.')

        for key, value in result.items():
            if key == 'mask':
                continue
            if (mask_value := self.mask_values.get(key, None)) is not None:
                value[~mask] = mask_value
            elif key not in classnames:
                value[~mask] = torch.nan

        return result


class BrainwebSlices(torch.utils.data.Dataset):
    """Dataset of 2D qMRI parameter slices based on Brainweb dataset.

    Dataset of agmented 2D qMRI parameter slices based on the segmentations of
    the Brainweb dataset [AubertBroche2006]_.


    References
    ----------
    .. [AubertBroche2006] Aubert-Broche, B., Griffin, M., Pike, G.B., Evans, A.C., & Collins, D.L. (2006).
       Twenty New Digital Brain Phantoms for Creation of Validation Image Data Bases.
       IEEE Transactions on Medical Imaging, 25 (11), 1410-1416. https://doi.org/10.1109/TMI.2006.883453
    """

    @staticmethod
    def download(
        output_directory: str | PathLike = CACHE_DIR_BRAINWEB,
        workers: int = 4,
        progress: bool = False,
        compress: bool = False,
    ) -> None:
        """Download Brainweb data.

        Parameters
        ----------
        output_directory
            Directory to save the data.
        workers
            Number of parallel downloads.
        progress
            Show progress bar.
        compress
            Use compression for HDF5 files. Saves disk space but might slow down or speed up access,
            depending on the file system and access pattern.

        """
        download_brainweb(output_directory, workers, progress, compress)

    def __init__(
        self,
        folder: str | Path = CACHE_DIR_BRAINWEB,
        what: Sequence[Literal['r1', 'r2', 'm0', 't1', 't2', 'mask', 'tissueclass'] | TClassNames] = ('m0', 'r1', 'r2'),
        parameters: Mapping[TClassNames, BrainwebTissue] = VALUES_3T_RANDOMIZED,
        orientation: Literal['axial', 'coronal', 'sagittal'] | Sequence[Literal['axial', 'coronal', 'sagittal']] = (
            'axial',
            'coronal',
            'sagittal',
        ),
        skip_slices: tuple[tuple[int, int], tuple[int, int], tuple[int, int]] = ((80, 80), (100, 100), (100, 100)),
        step: int = 1,
        slice_preparation: Callable[[torch.Tensor, torch.Generator | None], torch.Tensor] = DEFAULT_AUGMENT_256,
        mask_values: Mapping[str, float | None] = DEFAULT_VALUES,
        seed: int | Literal['index', 'random'] = 'random',
    ) -> None:
        """Initialize Brainweb qMRI slice phantom.

        Parameters
        ----------
        folder
            Folder with Brainweb data as HDF5 files.
        what
            What to return for each slice.
            Possible values are:
                - r1: R1 relaxation rate.
                - r2: R2 relaxation rate.
                - m0: M0 magnetization.
                - t1: T1 relaxation time.
                - t2: T2 relaxation time.
                - mask: mask indicating valid data.
                - tissueclass: class index.
        parameters
            Parameters for each tissue class.
            The Brainweb tissue classes are:
                - skl: skull
                - gry: gray matter
                - wht: white matter
                - csf: cerebrospinal fluid
                - mrw: bone marrow
                - dura: dura
                - fat: fat
                - fat2: fat and Tissue
                - mus: muscle
                - m-s: skin
                - ves: vessels
        orientation
            Orientation of slices (axial, coronal, or sagittal).
        skip_slices
            Specifies how much many slices to skip from the beginning and end for each of axial, coronal,
            or sagittal orientation.
        step
            Step size between slices, in voxel units (1 mm).
        slice_preparation
            Callable that performs slice augmentation and resizing, see `resize` or `augment` for examples.
            The default applies slight random rotation, shear, scaling, and flips, and scales to 256x256 images.
        mask_values
            Default values to use for masked out regions.
        seed
            Random seed. Can be an int, the strings ``index`` to use slice index as seed, or ``random`` for random seed.
        """
        self.parameters = parameters
        self.step = step
        self.what = what
        self.mask_values = mask_values

        if isinstance(orientation, str):
            orientation = [orientation]
        elif len(orientation) != len(set(orientation)):
            raise ValueError('Orientations must be unique.')
        try:
            self.axes = [{'axial': 0, 'coronal': 1, 'sagittal': 2}[o] for o in orientation]
        except KeyError:
            raise ValueError(f'Invalid axis: {orientation}.') from None

        self.skip_slices = skip_slices

        files_and_axes = []
        ns_slices = []
        h5_files = sorted(Path(folder).glob('s??.h5'))
        if not h5_files:
            raise FileNotFoundError(f'No files found in {folder}.')

        for axis in self.axes:
            skip_start, skip_end = self.skip_slices[axis]
            for fn in h5_files:
                with h5py.File(fn) as f:
                    n_slices = f['classes'].shape[axis] - skip_start - skip_end
                if n_slices > 0:
                    files_and_axes.append((fn, axis))
                    ns_slices.append(n_slices)
        if not files_and_axes:
            raise FileNotFoundError(f'After skipping {self.skip_slices} slices, no images are left.')

        self._files_and_axes = tuple(files_and_axes)
        self._ns_slices = np.cumsum([0, *ns_slices])

        self.slice_preparation = slice_preparation

        if seed == 'random':
            self._rng: torch.Generator | None = torch.default_generator
        elif isinstance(seed, int):
            self._rng = torch.Generator().manual_seed(seed)
        elif seed == 'index':
            self._rng = None

    def __len__(self) -> int:
        """Get number of slices."""
        return self._ns_slices[-1] // self.step

    def __getitem__(
        self, index: int
    ) -> dict[Literal['r1', 'r2', 'm0', 't1', 't2', 'mask', 'tissueclass'] | TClassNames, torch.Tensor]:
        """Get a single slice."""
        if index < 0:
            index = len(self) + index
        if not 0 <= index < len(self):
            raise IndexError
        index = index * self.step

        chunk_id = np.searchsorted(self._ns_slices, index, 'right') - 1
        file_path, axis = self._files_and_axes[chunk_id]
        slice_id = index - self._ns_slices[chunk_id] + self.skip_slices[axis][0]

        with h5py.File(file_path) as file:
            where = [slice(None)] * 3
            where[axis] = slice_id
            data = torch.as_tensor(np.array(file['classes'][tuple(where)], dtype=np.uint8))
            classnames = tuple(file.attrs['classnames'])

        rng = torch.Generator().manual_seed(index) if self._rng is None else self._rng
        data = self.slice_preparation(data.moveaxis(-1, 0) / 255, rng).moveaxis(
            0, -1
        )  # / 255 to convert from uint8 to 0...1
        mask = data.sum(-1) > 0.5
        result: dict[Literal['r1', 'r2', 'm0', 't1', 't2', 'mask', 'tissueclass'] | TClassNames, torch.Tensor] = {}

        for el in self.what:
            if el == 'r1':
                values = torch.stack([self.parameters[k].sample_r1(rng) for k in classnames])
                result[el] = (data.to(values) @ values)[..., 0]
            elif el == 'r2':
                values = torch.stack([self.parameters[k].sample_r2(rng) for k in classnames])
                result[el] = (data.to(values) @ values)[..., 0]
            elif el == 'm0':
                values = torch.stack([self.parameters[k].sample_m0(rng) for k in classnames])
                result[el] = (data.to(values) @ values)[..., 0]
            elif el == 't1':
                values = torch.stack([self.parameters[k].sample_r1(rng) for k in classnames])
                result[el] = (data.to(values) @ values)[..., 0].reciprocal()
            elif el == 't2':
                values = torch.stack([self.parameters[k].sample_r2(rng) for k in classnames])
                result[el] = (data.to(values) @ values)[..., 0].reciprocal()
            elif el == 'tissueclass':
                result[el] = data.argmax(-1)
            elif el in classnames:
                result[el] = data[..., classnames.index(el)]
            elif el == 'mask':
                result[el] = (
                    torch.nn.functional.conv2d((~mask)[None, None].float(), torch.ones(1, 1, 3, 3), padding=1)[0, 0] < 1
                )

            else:
                raise NotImplementedError(f'what=({el},) is not implemented.')

        for key, value in result.items():
            if key == 'mask':
                continue
            if (mask_value := self.mask_values.get(key, None)) is not None:
                value[~mask] = mask_value
            elif key not in classnames:
                value[~mask] = torch.nan

        return result



"""FastMRI dataset."""

from collections.abc import Callable, Sequence
from os import PathLike
from pathlib import Path
from warnings import warn

import h5py
import ismrmrd.xsd
import numpy as np
import torch
from einops import rearrange

from mrpro.algorithms.csm.inati import inati
from mrpro.data.AcqInfo import AcqInfo
from mrpro.data.KData import KData
from mrpro.data.KHeader import KHeader
from mrpro.data.traj_calculators.KTrajectoryCartesian import KTrajectoryCartesian
from mrpro.utils.interpolate import apply_lowres
from mrpro.utils.pad_or_crop import pad_or_crop
from mrpro.utils.reshape import unsqueeze_left


class FastMRIKDataDataset(torch.utils.data.Dataset):
    """FastMRI KData Dataset.

    This dataset returns KData objects for single slices or stacks of slices of the FastMRI brain or knee dataset.
    The data has to be downloaded beforehand. See https://fastmri.med.nyu.edu/ for more information.
    """

    def __init__(self, path: PathLike | str | Sequence[PathLike | str], single_slice: bool = True):
        """Initialize the dataset.

        Parameters
        ----------
        path : PathLike
            Either a path to a directory containing the FastMRI data as .h5 files or a sequence of paths of
            individual files.
        single_slice : bool
            Whether to return single slices or stacks of slices.
        """
        self._filenames = []
        slices = []
        for fn in Path(path).rglob('*.h5') if isinstance(path, str | Path | PathLike) else path:
            try:
                with h5py.File(fn, 'r') as file:
                    n_slices = file['kspace'].shape[0]
                    slices.append(n_slices)
                    self._filenames.append(fn)
            except (KeyError, FileNotFoundError, OSError):
                warn(f'Invalid file: {fn}. Skipping.', stacklevel=2)
        self._accum_slices = torch.tensor(slices).cumsum(dim=0) if single_slice else None

    def __len__(self) -> int:
        """Get length (number of slices or stacks of slices) of the dataset."""
        if self._accum_slices is None:
            return len(self._filenames)
        if len(self._accum_slices) == 0:
            return 0
        return int(self._accum_slices[-1])

    def __getitem__(self, idx: int) -> KData:
        """Get a single slice or stack of slices."""
        if not -len(self) <= idx < len(self):
            raise IndexError(f'Index {idx} is out of bounds for the dataset of size {len(self)}')
        if idx < 0:
            idx += len(self)
        if self._accum_slices is None:  # return stack
            file_idx = idx
            slice_idx: int | slice = slice(None)
        else:
            file_idx = int(torch.searchsorted(self._accum_slices, idx + 1))
            slice_idx = int(idx - self._accum_slices[file_idx])
        with h5py.File(self._filenames[file_idx], 'r') as file:
            # data is sometimes zero-padded, we remove the padding
            data = torch.as_tensor(np.array(file['kspace'][slice_idx]))
            data = unsqueeze_left(data, 4 - data.ndim)
            nonzero = data[0, 0, 0, :].abs() > 1e-12
            first, last = nonzero.nonzero().flatten()[[0, -1]]
            data = data[..., first : last + 1]
            data = rearrange(data, 'slices coils k0 k1 -> slices coils 1 k1 k0')
            n_k1, n_k0 = data.shape[-2:]
            info = AcqInfo()
            info.idx.k1 = torch.arange(first, last + 1)[None, None, None, :, None]
            header = KHeader.from_ismrmrd(
                ismrmrd.xsd.CreateFromDocument(file['ismrmrd_header'][()].decode('utf-8')), info
            )
            traj = KTrajectoryCartesian()(
                n_k0=n_k0,
                k0_center=n_k0 // 2,
                k1_idx=info.idx.k1,
                k1_center=first + n_k1 // 2,
                k2_idx=torch.tensor(0),
                k2_center=0,
            )
            kdata = KData(
                data=data,
                header=header,
                traj=traj,
            )
            return kdata


class FastMRIImageDataset(torch.utils.data.Dataset):
    """FastMRI Image Dataset.

    This dataset returns image tensors for single slices of the FastMRI brain or knee dataset.
    It filteres and resamples the files such that the returned images have a consistent shape
    of 320x320 before augmentations.

    The returned images are complex valued and will have shape ``(1, 1, 1, 320, 320)`` if coil combined or
    ``(1, n_coils, 1, 320, 320)`` otherwise.

    The data has to be downloaded beforehand. See https://fastmri.med.nyu.edu/ for more information.
    """

    def __init__(
        self,
        path: str | PathLike | Sequence[str | PathLike],
        coil_combine: bool = False,
        augment: Callable[[torch.Tensor, int], torch.Tensor] | None = None,
        allowed_n_coils: Sequence[int] | None = (16, 15),
    ):
        """Initialize the dataset.

        Parameters
        ----------
        path : PathLike
            Either a path to a directory containing the FastMRI data as .h5 files or a sequence of paths of
            individual files.
        coil_combine : bool
            Whether to perform coil combination sensitivity maps obtained using the Inati method.
            Note that this is **not** commonly used as the target for FastMRI challenges. Instead,
            as target the RSS combination of the coil images is used.
        augment
            Augmentation function. Will be called with the image and the index of the slices.
            If `coil_combine` is `True`, the function will be called with the complex valued coil combined image
            with shape (1, 320, 320) otherwise with the complex valued coil images with shape (n_coils, 320, 320).
            `None` means no augmentation.
        allowed_n_coils
            List of allowed number of coils. If `None`, all coils are allowed.
            The knee training set has 15 coils consistently, while the brain dataset has
            roughly 1300 files with 16 coils, 1100 files with 20 coils and 800 files with 4 coils.
            Only used if `coil_combine` is `False`.
        """
        slices = []
        self._filenames = []
        for fn in Path(path).rglob('*.h5') if isinstance(path, str | Path | PathLike) else path:
            try:
                with h5py.File(fn, 'r') as file:
                    acquisition: str = file.attrs['acquisition']
                    n_slices, n_coils = file['kspace'].shape[:2]
                    header = KHeader.from_ismrmrd(
                        ismrmrd.xsd.CreateFromDocument(file['ismrmrd_header'][()].decode('utf-8')), AcqInfo()
                    )
                    if acquisition.startswith('AX'):  # brain
                        if (
                            (not coil_combine and allowed_n_coils is not None and n_coils not in allowed_n_coils)
                            or header.recon_matrix.x not in (320, 384)
                            or round(header.recon_fov.y, 2) != 0.22
                        ):
                            continue
                    elif acquisition.endswith('FBK'):  # knee
                        if (
                            (not coil_combine and allowed_n_coils is not None and n_coils not in allowed_n_coils)
                            or header.recon_matrix.x != 320
                            or round(header.recon_fov.y, 2) != 0.14
                        ):
                            continue
                    else:
                        raise ValueError(f'Unknown acquisition: {acquisition}')
                    slices.append(n_slices)
                    self._filenames.append(fn)
            except (KeyError, FileNotFoundError, OSError):
                warn(f'Invalid file: {fn}. Skipping.', stacklevel=2)
        self._accum_slices = torch.tensor(slices).cumsum(dim=0)
        self._coil_combine = coil_combine
        self.augment = augment

    def __len__(self) -> int:
        """Get length (number of slices) of the dataset."""
        return int(self._accum_slices[-1])

    def __getitem__(self, idx: int) -> torch.Tensor:
        """Get a single slice."""
        if not -len(self) <= idx < len(self):
            raise IndexError(f'Index {idx} is out of bounds for the dataset of size {len(self)}')
        if idx < 0:
            idx += len(self)
        file_idx = torch.searchsorted(self._accum_slices, idx + 1)
        slice_idx = idx - self._accum_slices[file_idx]
        with h5py.File(self._filenames[file_idx], 'r') as file:
            data = torch.as_tensor(np.array(file['kspace'][slice_idx]))
            header = KHeader.from_ismrmrd(
                ismrmrd.xsd.CreateFromDocument(file['ismrmrd_header'][()].decode('utf-8')), AcqInfo()
            )
            # recon_matrix can be >320, but FOV is always constant. We crop in k-space to get the same resolution.
            data = pad_or_crop(
                data,
                (int(data.shape[-2] * 320 / header.recon_matrix.y), int(data.shape[-1] * 320 / header.recon_matrix.y)),
                dim=(-2, -1),
            )
            # we use an fft instead of ifft to already flip the image
            img = torch.fft.fftshift(torch.fft.fft2(torch.fft.fftshift(data, dim=(-2, -1))), dim=(-2, -1))
            # finally, we crop to the new recon_fov of 320x320 to remove oversampling
            img = pad_or_crop(img, (320, 320), dim=(-2, -1))
            if self._coil_combine:
                csm = apply_lowres(
                    lambda x: inati(x.unsqueeze(1), smoothing_width=5).squeeze(1), (32, 32), dim=(-2, -1)
                )(img)
                csm /= (csm * csm.conj()).sum(0, keepdim=True).sqrt()
                img = (img * csm.conj()).sum(dim=0, keepdim=True)
            if self.augment is not None:
                img = self.augment(img, idx)
            return rearrange(img, 'coils y x -> 1 coils 1 y x')



"""SpatialDimension dataclass."""

from __future__ import annotations

from collections.abc import Callable
from typing import Generic, cast, get_args

import numpy as np
import torch
from numpy.typing import ArrayLike
from typing_extensions import Protocol, Self, TypeVar, overload

import mrpro.utils.typing as type_utils
from mrpro.data.Dataclass import Dataclass
from mrpro.utils.indexing import Indexer

# Change here to add more types
VectorTypes = torch.Tensor
ScalarTypes = int | float
T = TypeVar('T', torch.Tensor, int, float)

# Covariant types, as SpatialDimension is a Container
# and we want, for example, SpatialDimension[int] to also be a SpatialDimension[float]
T_co = TypeVar('T_co', torch.Tensor, int, float, covariant=True)
T_co_float = TypeVar('T_co_float', float, torch.Tensor, covariant=True)
T_co_vector = torch.Tensor
T_co_scalar = TypeVar('T_co_scalar', int, float, covariant=True)


def _as_vectortype(x: ArrayLike) -> VectorTypes:
    """Convert ArrayLike to VectorType."""
    if isinstance(x, VectorTypes) and type(x) in get_args(VectorTypes):
        # exact type match
        return x
    if isinstance(x, VectorTypes):
        # subclass of torch.Tensor
        return torch.as_tensor(x)
    else:
        # any other ArrayLike (which is defined as convert to numpy array)
        return torch.as_tensor(np.asarray(x))


class XYZ(Protocol[T]):
    """Protocol for structures with attributes x, y and z of type T."""

    x: T
    y: T
    z: T


class SpatialDimension(Dataclass, Generic[T_co]):
    """Spatial dataclass of float/int/tensors (z, y, x)."""

    z: T_co
    y: T_co
    x: T_co

    @classmethod
    def from_xyz(cls, data: XYZ[T_co]) -> SpatialDimension[T_co]:
        """Create a SpatialDimension from something with (.x .y .z) parameters.

        Parameters
        ----------
        data
            should implement .x .y .z. For example ismrmrd's matrixSizeType.
        """
        return cls(data.z, data.y, data.x)

    @staticmethod
    def from_array_xyz(
        data: ArrayLike,
    ) -> SpatialDimension[torch.Tensor]:
        """Create a SpatialDimension from an arraylike interface.

        Parameters
        ----------
        data
            shape (..., 3) in the order (x,y,z)
        """
        data_ = _as_vectortype(data)
        if np.size(data_, -1) != 3:
            raise ValueError(f'Expected last dimension to be 3, got {np.size(data_, -1)}')

        x = data_[..., 0]
        y = data_[..., 1]
        z = data_[..., 2]

        return SpatialDimension(z, y, x)

    @staticmethod
    def from_array_zyx(
        data: ArrayLike,
    ) -> SpatialDimension[torch.Tensor]:
        """Create a SpatialDimension from an arraylike interface.

        Parameters
        ----------
        data
            shape (..., 3) in the order (z,y,x)
        """
        data_ = _as_vectortype(data)
        if np.size(data_, -1) != 3:
            raise ValueError(f'Expected last dimension to be 3, got {np.size(data_, -1)}')

        x = data_[..., 2]
        y = data_[..., 1]
        z = data_[..., 0]

        return SpatialDimension(z, y, x)

    # This function is mainly for type hinting and docstring
    def apply_(self, function: Callable[[T], T] | None = None, **_) -> Self:
        """Apply a function to each z, y, x (in-place).

        Parameters
        ----------
        function
            function to apply
        """
        return super().apply_(function)

    # This function is mainly for type hinting and docstring
    def apply(self, function: Callable[[T], T] | None = None, **_) -> Self:
        """Apply a function to each z, y, x (returning a new object).

        Parameters
        ----------
        function
            function to apply
        """
        return super().apply(function)

    @property
    def zyx(self) -> tuple[T_co, T_co, T_co]:
        """Return a z,y,x tuple.

        ```{note}
        To access the z, y, x values, use the attributes directly.
        ```
        """
        return (self.z, self.y, self.x)

    def __shortstr__(self) -> str:
        """Return a short string representation."""
        if isinstance(self.x, VectorTypes) or isinstance(self.y, VectorTypes) or isinstance(self.z, VectorTypes):
            super().__shortstr__()
        return f'z={self.z}, y={self.y}, x={self.x}'

    def __getitem__(self: SpatialDimension[T_co], idx: type_utils.TorchIndexerType | Indexer) -> SpatialDimension[T_co]:
        """Get SpatialDimension item."""
        if not all(isinstance(el, VectorTypes) for el in self.zyx):
            return self
        return cast(SpatialDimension[T_co], super().__getitem__(idx))

    def __setitem__(self: SpatialDimension[T_co_vector], idx: type_utils.TorchIndexerType, other: SpatialDimension):
        """Set SpatialDimension item."""
        if not all(isinstance(el, VectorTypes) for el in self.zyx):
            raise IndexError('Cannot index SpatialDimension with non-indexable members')
        self.z[idx] = other.z
        self.y[idx] = other.y
        self.x[idx] = other.x

    @overload
    def __mul__(self: SpatialDimension[T_co], other: T_co | SpatialDimension[T_co]) -> SpatialDimension[T_co]: ...

    @overload
    def __mul__(self: SpatialDimension, other: SpatialDimension[T_co_vector]) -> SpatialDimension[T_co_vector]: ...

    @overload
    def __mul__(self: SpatialDimension[int], other: float | SpatialDimension[float]) -> SpatialDimension[float]: ...

    @overload
    def __mul__(
        self: SpatialDimension[T_co_float], other: float | SpatialDimension[float]
    ) -> SpatialDimension[T_co_float]: ...

    def __mul__(self: SpatialDimension[T_co], other: float | T_co | SpatialDimension) -> SpatialDimension:
        """Multiply SpatialDimension with numeric other or SpatialDimension."""
        if isinstance(other, SpatialDimension):
            return SpatialDimension(self.z * other.z, self.y * other.y, self.x * other.x)
        return SpatialDimension(self.z * other, self.y * other, self.x * other)

    @overload
    def __rmul__(self: SpatialDimension[T_co], other: T_co) -> SpatialDimension[T_co]: ...
    @overload
    def __rmul__(self: SpatialDimension[int], other: float) -> SpatialDimension[float]: ...

    @overload
    def __rmul__(self: SpatialDimension[T_co_float], other: float) -> SpatialDimension[T_co_float]: ...

    def __rmul__(self: SpatialDimension[T_co], other: float | T_co | SpatialDimension) -> SpatialDimension:
        """Right multiply SpatialDimension with numeric other or SpatialDimension."""
        return self.__mul__(other)

    @overload
    def __truediv__(self: SpatialDimension[int], other: float | SpatialDimension[float]) -> SpatialDimension[float]: ...

    @overload
    def __truediv__(self: SpatialDimension, other: SpatialDimension[T_co_vector]) -> SpatialDimension[T_co_vector]: ...

    @overload
    def __truediv__(self: SpatialDimension[T_co], other: T_co | SpatialDimension[T_co]) -> SpatialDimension[T_co]: ...

    @overload
    def __truediv__(
        self: SpatialDimension[T_co_float], other: float | SpatialDimension[float]
    ) -> SpatialDimension[T_co_float]: ...

    def __truediv__(self: SpatialDimension[T_co], other: float | T_co | SpatialDimension) -> SpatialDimension:
        """Divide SpatialDimension by numeric other or SpatialDimension."""
        if isinstance(other, SpatialDimension):
            return SpatialDimension(self.z / other.z, self.y / other.y, self.x / other.x)
        return SpatialDimension(self.z / other, self.y / other, self.x / other)

    @overload
    def __rtruediv__(self: SpatialDimension[int], other: float) -> SpatialDimension[float]: ...
    @overload
    def __rtruediv__(self: SpatialDimension[T_co], other: T_co) -> SpatialDimension[T_co]: ...

    @overload
    def __rtruediv__(self: SpatialDimension[T_co_float], other: float) -> SpatialDimension[T_co_float]: ...

    def __rtruediv__(self: SpatialDimension[T_co], other: float | T_co) -> SpatialDimension:
        """Divide SpatialDimension or numeric other by SpatialDimension."""
        return SpatialDimension(other / self.z, other / self.y, other / self.x)

    @overload
    def __add__(self: SpatialDimension[T_co], other: T_co | SpatialDimension[T_co]) -> SpatialDimension[T_co]: ...

    @overload
    def __add__(self: SpatialDimension, other: SpatialDimension[T_co_vector]) -> SpatialDimension[T_co_vector]: ...

    @overload
    def __add__(self: SpatialDimension[int], other: float | SpatialDimension[float]) -> SpatialDimension[float]: ...

    @overload
    def __add__(
        self: SpatialDimension[T_co_float], other: float | SpatialDimension[float]
    ) -> SpatialDimension[T_co_float]: ...

    def __add__(self: SpatialDimension[T_co], other: float | T_co | SpatialDimension) -> SpatialDimension:
        """Add SpatialDimension or numeric other to SpatialDimension."""
        if isinstance(other, SpatialDimension):
            return SpatialDimension(self.z + other.z, self.y + other.y, self.x + other.x)
        return SpatialDimension(self.z + other, self.y + other, self.x + other)

    @overload
    def __radd__(self: SpatialDimension[T_co], other: T_co) -> SpatialDimension[T_co]: ...

    @overload
    def __radd__(self: SpatialDimension[int], other: float) -> SpatialDimension[float]: ...

    @overload
    def __radd__(self: SpatialDimension[T_co_float], other: float) -> SpatialDimension[T_co_float]: ...

    def __radd__(self: SpatialDimension[T_co], other: float | T_co) -> SpatialDimension:
        """Right add numeric other to SpatialDimension."""
        return self.__add__(other)

    @overload
    def __floordiv__(self: SpatialDimension[T_co], other: T_co | SpatialDimension[T_co]) -> SpatialDimension[T_co]: ...

    @overload
    def __floordiv__(
        self: SpatialDimension[int], other: float | SpatialDimension[float]
    ) -> SpatialDimension[float]: ...

    @overload
    def __floordiv__(self: SpatialDimension, other: SpatialDimension[T_co_vector]) -> SpatialDimension[T_co_vector]: ...

    @overload
    def __floordiv__(
        self: SpatialDimension[T_co_float], other: float | SpatialDimension[float]
    ) -> SpatialDimension[T_co_float]: ...

    def __floordiv__(self: SpatialDimension[T_co], other: float | T_co | SpatialDimension) -> SpatialDimension:
        """Floor divide SpatialDimension by numeric other."""
        if isinstance(other, SpatialDimension):
            return SpatialDimension(self.z // other.z, self.y // other.y, self.x // other.x)
        return SpatialDimension(self.z // other, self.y // other, self.x // other)

    @overload
    def __rfloordiv__(self: SpatialDimension[T_co], other: T_co) -> SpatialDimension[T_co]: ...

    @overload
    def __rfloordiv__(self: SpatialDimension[int], other: float) -> SpatialDimension[float]: ...

    @overload
    def __rfloordiv__(self: SpatialDimension[T_co_float], other: float) -> SpatialDimension[T_co_float]: ...

    def __rfloordiv__(self: SpatialDimension[T_co], other: float | T_co) -> SpatialDimension:
        """Floor divide other by SpatialDimension."""
        if isinstance(other, SpatialDimension):
            return SpatialDimension(other.z // self.z, other.y // self.y, other.x // self.x)
        return SpatialDimension(other // self.z, other // self.y, other // self.x)

    @overload
    def __sub__(self: SpatialDimension[T_co], other: T_co | SpatialDimension[T_co]) -> SpatialDimension[T_co]: ...

    @overload
    def __sub__(self: SpatialDimension, other: SpatialDimension[T_co_vector]) -> SpatialDimension[T_co_vector]: ...

    @overload
    def __sub__(self: SpatialDimension[int], other: float | SpatialDimension[float]) -> SpatialDimension[float]: ...

    @overload
    def __sub__(
        self: SpatialDimension[T_co_float], other: float | SpatialDimension[float]
    ) -> SpatialDimension[T_co_float]: ...

    def __sub__(self: SpatialDimension[T_co], other: float | T_co | SpatialDimension) -> SpatialDimension:
        """Subtract SpatialDimension or numeric other to SpatialDimension."""
        if isinstance(other, SpatialDimension):
            return SpatialDimension(self.z - other.z, self.y - other.y, self.x - other.x)
        return SpatialDimension(self.z - other, self.y - other, self.x - other)

    @overload
    def __rsub__(self: SpatialDimension[T_co], other: T_co) -> SpatialDimension[T_co]: ...

    @overload
    def __rsub__(self: SpatialDimension[int], other: float) -> SpatialDimension[float]: ...

    @overload
    def __rsub__(self: SpatialDimension[T_co_float], other: float) -> SpatialDimension[T_co_float]: ...

    def __rsub__(self: SpatialDimension[T_co], other: float | T_co) -> SpatialDimension:
        """Subtract SpatialDimension from numeric other or SpatialDimension."""
        if isinstance(other, SpatialDimension):
            return SpatialDimension(other.z - self.z, other.y - self.y, other.x - self.x)
        return SpatialDimension(other - self.z, other - self.y, other - self.x)

    def __neg__(self: SpatialDimension[T_co]) -> SpatialDimension[T_co]:
        """Negate SpatialDimension."""
        return SpatialDimension(-self.z, -self.y, -self.x)

    @overload
    def __eq__(self: SpatialDimension[T_co_scalar], other: object) -> bool: ...
    @overload
    def __eq__(self: SpatialDimension[T_co_vector], other: SpatialDimension[T_co_vector]) -> T_co_vector: ...

    def __eq__(
        self: SpatialDimension[T_co_scalar] | SpatialDimension[T_co_vector],
        other: object | SpatialDimension[T_co_vector],
    ) -> bool | T_co_vector:
        """Check if self is equal to other."""
        if not isinstance(other, SpatialDimension):
            return NotImplemented
        return (self.z == other.z) & (self.y == other.y) & (self.x == other.x)

    @overload
    def __lt__(self: SpatialDimension[T_co_vector], other: SpatialDimension[T_co_vector]) -> T_co_vector: ...
    @overload
    def __lt__(self: SpatialDimension[T_co_scalar], other: SpatialDimension[T_co_scalar]) -> bool: ...
    def __lt__(
        self: SpatialDimension[T_co_scalar] | SpatialDimension[T_co_vector],
        other: SpatialDimension[T_co_scalar] | SpatialDimension[T_co_vector],
    ) -> bool | T_co_vector:
        """Check if self is less than other."""
        if not isinstance(other, SpatialDimension):
            return NotImplemented
        return (self.x < other.x) & (self.y < other.y) & (self.z < other.z)

    @overload
    def __le__(self: SpatialDimension[T_co_vector], other: SpatialDimension[T_co_vector]) -> T_co_vector: ...
    @overload
    def __le__(self: SpatialDimension[T_co_scalar], other: SpatialDimension[T_co_scalar]) -> bool: ...
    def __le__(
        self: SpatialDimension[T_co_scalar] | SpatialDimension[T_co_vector],
        other: SpatialDimension[T_co_scalar] | SpatialDimension[T_co_vector],
    ) -> bool | T_co_vector:
        """Check if self is less than or equal to other."""
        if not isinstance(other, SpatialDimension):
            return NotImplemented
        return (self.x <= other.x) & (self.y <= other.y) & (self.z <= other.z)

    @overload
    def __gt__(self: SpatialDimension[T_co_vector], other: SpatialDimension[T_co_vector]) -> T_co_vector: ...
    @overload
    def __gt__(self: SpatialDimension[T_co_scalar], other: SpatialDimension[T_co_scalar]) -> bool: ...
    def __gt__(
        self: SpatialDimension[T_co_scalar] | SpatialDimension[T_co_vector],
        other: SpatialDimension[T_co_scalar] | SpatialDimension[T_co_vector],
    ) -> bool | T_co_vector:
        """Check if self is greater than other."""
        if not isinstance(other, SpatialDimension):
            return NotImplemented
        return (self.x > other.x) & (self.y > other.y) & (self.z > other.z)

    @overload
    def __ge__(self: SpatialDimension[T_co_vector], other: SpatialDimension[T_co_vector]) -> T_co_vector: ...
    @overload
    def __ge__(self: SpatialDimension[T_co_scalar], other: SpatialDimension[T_co_scalar]) -> bool: ...
    def __ge__(
        self: SpatialDimension[T_co_scalar] | SpatialDimension[T_co_vector],
        other: SpatialDimension[T_co_scalar] | SpatialDimension[T_co_vector],
    ) -> bool | T_co_vector:
        """Check if self is greater than or equal to other."""
        if not isinstance(other, SpatialDimension):
            return NotImplemented
        return (self.x >= other.x) & (self.y >= other.y) & (self.z >= other.z)

    def __post_init__(self) -> None:
        """Ensure that the data is of matching shape."""
        if not all(isinstance(val, (int | float)) for val in self.zyx):
            self.z, self.y, self.x = (_as_vectortype(v) for v in self.zyx)  # type: ignore[assignment]
            try:
                torch.broadcast_tensors(self.z, self.y, self.x)
            except RuntimeError:
                raise ValueError('The shapes of the tensors are not broadcastable') from None

    @property
    def shape(self) -> torch.Size:
        """Get the (broadcasted) shape.

        Returns
        -------
            Empty tuple if x, y, and z are scalar types, otherwise shape.
            The shape is the smallest shape x,y, and z can be broadcasted to.
            None of the fields need to have this shape.
        """
        if isinstance(self.x, ScalarTypes) and isinstance(self.y, ScalarTypes) and isinstance(self.z, ScalarTypes):
            return torch.Size()
        else:
            return super().shape

    # Note: mypy doesnt't support overloads on properties yet
    @property
    def dtype(self) -> torch.dtype | type[int] | type[float]:
        """Get the (promoted) dtype of the fields."""
        if isinstance(self.x, VectorTypes) and isinstance(self.y, VectorTypes) and isinstance(self.z, VectorTypes):
            return torch.promote_types(torch.promote_types(self.x.dtype, self.y.dtype), self.z.dtype)
        elif isinstance(self.x, int) and isinstance(self.y, int) and isinstance(self.z, int):
            return int
        elif isinstance(self.x, int | float) and isinstance(self.y, int | float) and isinstance(self.z, int | float):
            return float
        else:
            raise TypeError('Inconsistent types of the fields.')



from mrpro.operators.functionals.L1Norm import L1Norm
from mrpro.operators.functionals.L1NormViewAsReal import L1NormViewAsReal
from mrpro.operators.functionals.L2NormSquared import L2NormSquared
from mrpro.operators.functionals.MSE import MSE
from mrpro.operators.functionals.SSIM import SSIM
from mrpro.operators.functionals.ZeroFunctional import ZeroFunctional
__all__ = ["L1Norm", "L1NormViewAsReal", "L2NormSquared", "MSE", "SSIM","ZeroFunctional"]



"""Saturation recovery signal model for T1 mapping."""

from collections.abc import Sequence

import torch

from mrpro.operators.SignalModel import SignalModel
from mrpro.utils.reshape import unsqueeze_right


class SaturationRecovery(SignalModel[torch.Tensor, torch.Tensor]):
    """Signal model for saturation recovery."""

    def __init__(self, saturation_time: float | torch.Tensor | Sequence[float]) -> None:
        """Initialize saturation recovery signal model for T1 mapping.

        Parameters
        ----------
        saturation_time
            delay between saturation and acquisition. Shape `(time, ...)`.
        """
        super().__init__()
        saturation_time = torch.as_tensor(saturation_time)
        self.saturation_time = torch.nn.Parameter(saturation_time, requires_grad=saturation_time.requires_grad)

    def __call__(self, m0: torch.Tensor, t1: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply the saturation recovery signal model.

        Calculates the signal based on the formula:
        :math:`S(t_{sat}) = M_0 (1 - e^{-t_{sat} / T_1})`,
        where `t_{sat}` are the saturation times.

        Parameters
        ----------
        m0
            Equilibrium signal / proton density.
            Shape `(...)`, for example `(*other, coils, z, y, x)` or `(samples)`.
        t1
            Longitudinal relaxation time T1.
            Shape `(...)`, for example `(*other, coils, z, y, x)` or `(samples)`.

        Returns
        -------
            Signal calculated for each saturation time.
            Shape `(times ...)`, for example `(times, *other, coils, z, y, x)`, or `(times, samples)`
            where `times` is the number of saturation times.
        """
        return super().__call__(m0, t1)

    def forward(self, m0: torch.Tensor, t1: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward of SaturationRecovery.

        .. note::
            Prefer calling the instance of the SaturationRecovery as ``operator(x)`` over directly calling this method.
            See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        ndim = max(m0.ndim, t1.ndim)
        saturation_time = unsqueeze_right(
            self.saturation_time, ndim - self.saturation_time.ndim + 1
        )  # leftmost is time
        signal = m0 * (1 - torch.exp(-(saturation_time / t1)))
        return (signal,)



"""Slice Profiles."""

import abc
from collections.abc import Sequence
from math import log

import numpy as np
import torch

from mrpro.utils.TensorAttributeMixin import TensorAttributeMixin

__all__ = ['SliceGaussian', 'SliceInterpolate', 'SliceProfileBase', 'SliceSmoothedRectangular']


class SliceProfileBase(abc.ABC, TensorAttributeMixin, torch.nn.Module):
    """Base class for slice profiles."""

    @abc.abstractmethod
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Evaluate the slice profile at a position."""
        raise NotImplementedError

    def random_sample(self, size: Sequence[int]) -> torch.Tensor:
        """Sample `n` random positions from the profile.

        Use the profile as a probability density function to sample positions.

        Parameters
        ----------
        size
            Number of positions to sample

        Returns
        -------
            Sampled positions, shape will be size.
        """
        raise NotImplementedError


class SliceGaussian(SliceProfileBase):
    """Gaussian slice profile."""

    fwhm: torch.Tensor

    def __init__(self, fwhm: float | torch.Tensor):
        """Initialize the Gaussian slice profile.

        Parameters
        ----------
        fwhm
            Full width at half maximum of the Gaussian
        """
        super().__init__()
        self.fwhm = torch.as_tensor(fwhm)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Evaluate the Gaussian slice profile at a position.

        Parameters
        ----------
        x
            Position at which to evaluate the profile

        Returns
        -------
            Value of the profile / intensity at the given position
        """
        return torch.exp(-(x**2) / (0.36 * self.fwhm**2))


class SliceSmoothedRectangular(SliceProfileBase):
    """Rectangular slice profile with smoothed flanks.

    Implemented as a convolution of a rectangular profile
    with a Gaussian.
    """

    def __init__(self, fwhm_rect: float | torch.Tensor, fwhm_gauss: float | torch.Tensor):
        """Initialize the Rectangular slice profile.

        Parameters
        ----------
        fwhm_rect
            Full width at half maximum of the rectangular profile
        fwhm_gauss
            Full width at half maximum of the Gaussian profile.
            Set to zero to disable smoothing.

        Returns
        -------
            Value of the profile / intensity at the given position
        """
        super().__init__()
        self.fwhm_rect: torch.Tensor
        self.fwhm_gauss: torch.Tensor

        self.register_buffer('fwhm_rect', torch.as_tensor(fwhm_rect))
        self.register_buffer('fwhm_gauss', torch.as_tensor(fwhm_gauss))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Evaluate the Gaussian slice profile at a position.

        Parameters
        ----------
        x
            Position at which to evaluate the profile

        Returns
        -------
            Value of the profile / intensity at the given position
        """
        scaled = x * 2 / self.fwhm_rect
        if self.fwhm_gauss > 0 and self.fwhm_rect > 0:
            n = (log(2) ** 0.5) * self.fwhm_rect / self.fwhm_gauss
            norm = 1 / (2 * torch.erf(n))
            return (torch.erf(n * (1 - scaled)) + torch.erf(n * (1 + scaled))) * norm
        elif self.fwhm_rect > 0:
            return (scaled.abs() <= 1).float()
        elif self.fwhm_gauss > 0:
            return torch.exp(-4 * log(2) * (x / self.fwhm_gauss) ** 2)
        else:
            raise ValueError('At least one of the widths has to be greater zero.')


class SliceInterpolate(SliceProfileBase):
    """slice profile based on interpolation of measured profile."""

    def __init__(self, positions: torch.Tensor, values: torch.Tensor):
        """Initialize the interpolated slice profile.

        Parameters
        ----------
        positions
            Positions of the measured profile.
        values
            Intensities of the measured profile.
        """
        super().__init__()
        self._xs = positions.detach().cpu().float().numpy()
        self._weights = values.detach().cpu().float().numpy()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Evaluate the interpolated slice profile at a position.

        Parameters
        ----------
        x
            Position at which to evaluate the profile.

        Returns
        -------
            Value of the profile / intensity at the given position.
        """
        if x.requires_grad:
            raise NotImplementedError('Interpolated profile does not support gradients.')
        x_np = x.detach().cpu().numpy()
        y_np = torch.as_tensor(np.interp(x_np, self._xs, self._weights, 0, 0))
        return y_np.to(x.device)



"""Class for Density Compensation Operator."""

import torch

from mrpro.data.DcfData import DcfData
from mrpro.operators.EinsumOp import EinsumOp


class DensityCompensationOp(EinsumOp):
    """Density Compensation Operator."""

    def __init__(self, dcf: DcfData | torch.Tensor) -> None:
        """Initialize a Density Compensation Operator.

        Parameters
        ----------
        dcf
           Density Compensation Data
        """
        if isinstance(dcf, DcfData):
            # only tensors can currently be used as buffers
            # thus, einsumop is initialized with the tensor data
            # TODO: change if einsumop can handle dataclasses
            dcf_tensor = dcf.data
        else:
            dcf_tensor = dcf
        super().__init__(dcf_tensor, '...,... -> ...')

    def __call__(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply density compensation to k-space data.

        This operator performs an element-wise multiplication of the input k-space data
        with the density compensation factors (DCF).

        Parameters
        ----------
        x
            Input k-space data.

        Returns
        -------
            Density compensated k-space data.
        """
        return super().__call__(x)



"""Class for 3D->2D Projection Operator."""

import itertools
import warnings
from collections.abc import Callable, Sequence
from typing import Literal, TypeAlias, cast

import einops
import numpy as np
import torch
from torch import Tensor

from mrpro.data.Rotation import Rotation
from mrpro.data.SpatialDimension import SpatialDimension
from mrpro.operators.LinearOperator import LinearOperator
from mrpro.utils.reshape import ravel_multi_index
from mrpro.utils.slice_profiles import SliceSmoothedRectangular
from mrpro.utils.typing import NestedSequence


class _MatrixMultiplicationCtx(torch.autograd.function.FunctionCtx):
    """Autograd context for matrix multiplication, used for type hinting."""

    x_is_complex: bool
    saved_tensors: tuple[Tensor]


class _MatrixMultiplication(torch.autograd.Function):
    """Helper for matrix multiplication.

    For sparse matrices, it can be more efficient to have a
    separate representation of the adjoint matrix to be used
    in the backward pass.

    """

    @staticmethod
    def forward(x: Tensor, matrix: Tensor, matrix_adjoint: Tensor) -> Tensor:  # noqa: ARG004
        if x.is_complex() == matrix.is_complex():
            return matrix @ x
        # required for sparse matrices to support mixed complex/real multiplication
        elif x.is_complex():
            return torch.complex(matrix @ x.real, matrix @ x.imag)
        else:
            return torch.complex(matrix.real @ x, matrix.imag @ x)

    @staticmethod
    def setup_context(
        ctx: _MatrixMultiplicationCtx,
        inputs: tuple[Tensor, Tensor, Tensor],
        outputs: tuple[Tensor],  # noqa: ARG004
    ) -> None:
        x, _, matrix_adjoint = inputs
        ctx.x_is_complex = x.is_complex()
        ctx.save_for_backward(matrix_adjoint)

    @staticmethod
    def backward(ctx: _MatrixMultiplicationCtx, *grad_output: Tensor) -> tuple[Tensor, None, None]:
        (matrix_adjoint,) = ctx.saved_tensors
        if ctx.x_is_complex:
            if matrix_adjoint.is_complex() == grad_output[0].is_complex():
                grad_x = matrix_adjoint @ grad_output[0]
            elif matrix_adjoint.is_complex():
                grad_x = torch.complex(matrix_adjoint.real @ grad_output[0], matrix_adjoint.imag @ grad_output[0])
            else:
                grad_x = torch.complex(matrix_adjoint @ grad_output[0].real, matrix_adjoint @ grad_output[0].imag)
        else:  # real grad
            grad_x = matrix_adjoint.real @ grad_output[0].real
            if matrix_adjoint.is_complex() and grad_output[0].is_complex():
                grad_x -= matrix_adjoint.imag @ grad_output[0].imag
        return grad_x, None, None


TensorFunction: TypeAlias = Callable[[Tensor], Tensor]


class SliceProjectionOp(LinearOperator):
    """Slice Projection Operator.

    This operation samples from a 3D Volume a slice with a given rotation and shift
    (relative to the center of the volume) according to the `slice_profile`.
    It can, for example, be used to describe the slice selection of a 2D MRI sequence
    from the 3D Volume.

    The projection will be done by sparse matrix multiplication.

    `slice_rotation`, `slice_shift`, and `slice_profile` can have (multiple) batch dimensions. These dimensions will
    be broadcasted to a common shape and added to the front of the volume.
    Different settings for different volume batches are NOT supported, consider creating multiple
    operators if required.
    """

    matrix: Tensor | None
    matrix_adjoint: Tensor | None

    def __init__(
        self,
        input_shape: SpatialDimension[int],
        slice_rotation: Rotation | None = None,
        slice_shift: float | Tensor = 0.0,
        slice_profile: TensorFunction | np.ndarray | NestedSequence[TensorFunction] | float = 2.0,
        optimize_for: Literal['forward', 'adjoint', 'both'] = 'both',
    ):
        """Create a module that represents the 'projection' of a volume onto a plane.

        This operation samples from a 3D Volume a slice with a given rotation and shift
        (relative to the center of the volume) according to the slice_profile.
        It can, for example, be used to describe the slice selection of a 2D MRI sequence
        from the 3D Volume.

        The projection will be done by sparse matrix multiplication.

        Rotation, shift, and profile can have (multiple) batch dimensions. These dimensions will
        be broadcasted to a common shape and added to the front of the volume.
        Different settings for different volume batches are NOT supported, consider creating multiple
        operators if required.

        Note
        ----
        All parameters must be on cpu when creating the operator. Preparation of the operator on
        the GPU would be slower than transferring the parameters to the CPU, creation, then transferring
        the operator back to the GPU.


        Parameters
        ----------
        input_shape
            Shape of the 3D volume to sample from. `(z, y, x)`
        slice_rotation
            Rotation that describes the orientation of the plane. If `None`,
            an identity rotation is used.
        slice_shift
            Offset of the plane in the volume perpendicular plane from the center of the volume.
            (The center of a 4 pixel volume is between 1 and 2.)
        slice_profile
            A function returning the relative intensity of the slice profile at a position x
            (relative to the nominal profile center). This can also be a nested Sequence or an
            numpy array of functions. See `mrpro.utils.slice_profiles` for examples.
            If it is a single float, it will be interpreted as the full-width-at-half-maximum (FWHM) of a rectangular
            profile.
        optimize_for
            Whether to optimize for forward or adjoint operation or both.
            Optimizing for both takes more memory but is faster for both operations.

        """
        super().__init__()
        if isinstance(slice_profile, float | int):
            slice_profile = SliceSmoothedRectangular(slice_profile, 0.0)
        slice_profile_array = np.array(slice_profile)

        if slice_rotation is None:
            slice_rotation = Rotation.identity()
        elif slice_rotation.device.type != 'cpu':
            raise ValueError('slice_rotation must be on cpu')

        max_shape = max(input_shape.z, input_shape.y, input_shape.x)

        def _find_width(slice_profile: TensorFunction) -> int:
            # figure out how far along the profile we have to consider values
            # clip up to 0.01 of intensity on both sides
            test_values = torch.arange(-max_shape, max_shape, max_shape)
            profile = slice_profile(test_values)
            cdf = torch.cumsum(profile, -1) / profile.sum()
            left = test_values[(cdf > 0.01).int().argmax()]
            right = test_values[(cdf > 0.99).int().argmax()]
            return int(max(left.abs().item(), right.abs().item())) + 1

        def _at_least_width_1(slice_profile: TensorFunction):
            test_values = torch.linspace(-0.5, 0.5, 100)
            return (slice_profile(test_values) > 1e-6).all()

        if not np.vectorize(_at_least_width_1)(slice_profile_array).all():
            raise ValueError(
                'The slice profile must have a width of at least 1 voxel,'
                ' i.e. the profile should be greater then 1e-6 in (-0.5, 0.5)'
            )

        slice_shift_tensor = cast(torch.Tensor, torch.atleast_1d(torch.as_tensor(slice_shift)))
        if slice_shift_tensor.device.type != 'cpu':
            raise ValueError('slice_shift must be on cpu')
        batch_shapes = torch.broadcast_shapes(slice_rotation.shape, slice_shift_tensor.shape, slice_profile_array.shape)
        assert isinstance(batch_shapes, torch.Size)  # noqa: S101 # mypy
        rotation_quats = torch.broadcast_to(slice_rotation.as_quat(), (*batch_shapes, 4)).reshape(-1, 4)
        slice_rotation = Rotation(rotation_quats, normalize=False, copy=False)
        slice_shift_tensor = torch.broadcast_to(slice_shift_tensor, batch_shapes).flatten()
        widths = np.broadcast_to(np.vectorize(_find_width)(slice_profile_array), batch_shapes).ravel()
        slice_profile_array = np.broadcast_to(slice_profile_array, batch_shapes).ravel()
        matrices = [
            SliceProjectionOp.projection_matrix(
                input_shape,
                SpatialDimension(1, max_shape, max_shape),
                offset=torch.tensor([shift, 0.0, 0.0]),
                slice_function=f,
                rotation=rot,
                w=int(w),
            )
            for rot, shift, f, w in zip(slice_rotation, slice_shift_tensor, slice_profile_array, widths, strict=True)
        ]
        matrix = SliceProjectionOp.join_matrices(matrices)

        # in csr format the matmul is faster, but saving one for forward and adjoint takes more memory
        with warnings.catch_warnings():
            # beta status in pytorch causes a warning to be printed
            warnings.filterwarnings('ignore', category=UserWarning, message='Sparse')
            if optimize_for == 'forward':
                self.matrix = matrix.to_sparse_csr()
                self.matrix_adjoint = None
            elif optimize_for == 'adjoint':
                self.matrix_adjoint = matrix.H.to_sparse_csr()
                self.matrix = None
            elif optimize_for == 'both':
                self.matrix_adjoint = matrix.H.to_sparse_csr()
                self.matrix = matrix.to_sparse_csr()

            else:
                raise ValueError("optimize_for must be one of 'forward', 'adjoint', 'both'")

        self._range_shape: tuple[int] = (*batch_shapes, 1, max_shape, max_shape)
        self._domain_shape = input_shape.zyx

    def __call__(self, x: Tensor) -> tuple[Tensor]:
        """Transform a 3D volume to a 2D slice.

        This operation uses a sparse matrix multiplication, where the matrix
        is determined by the slice rotation, shift, and profile defined
        at initialization. It samples the input 3D volume to generate a 2D slice.

        Parameters
        ----------
        x
            Input 3D Volume tensor. Expected shape `(..., Z, Y, X)`,
            where Z, Y, X match the `input_shape` provided during initialization.
            `(...)` represents optional batch dimensions.

        Returns
        -------
            The 2D slice. The shape of the slice will be
            `(..., 1, max_dim, max_dim)` where `max_dim` is the maximum of
            the input Z, Y, X dimensions, and `(...)` matches input batch dimensions.
        """
        return super().__call__(x)

    def forward(self, x: Tensor) -> tuple[Tensor]:
        """Apply forward of SliceProjectionOp.

        .. note::
            Prefer calling the instance of the SliceProjectionOp operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        match (self.matrix, self.matrix_adjoint):
            # selection based on the optimize_for setting
            case (None, None):
                raise RuntimeError('Either matrix or matrix adjoint must be set')
            case (matrix, None) if matrix is not None:
                matrix_adjoint = matrix.H
            case (None, matrix_adjoint) if matrix_adjoint is not None:
                matrix = matrix_adjoint.H
            case (matrix, matrix_adjoint):
                ...

        # For the (unusual case) of batched volumes, we will apply for each element in series
        xflat = torch.atleast_2d(einops.rearrange(x, '... x y z -> (...) (x y z)'))
        yl = [_MatrixMultiplication.apply(x, matrix, matrix_adjoint) for x in xflat]

        y = torch.stack([el.reshape(self._range_shape) for el in yl], -4)
        y = y.reshape(*y.shape[:-4], *x.shape[:-3], *y.shape[-3:])
        return (y,)

    def adjoint(self, x: Tensor) -> tuple[Tensor,]:
        """Transform a 2D slice back into a 3D volume.

        This operation uses the adjoint of the sparse projection matrix
        to map the input 2D slice data back into the 3D volume.

        Parameters
        ----------
        x
            Input 2D Slice tensor. Expected shape is `(..., 1, max_dim, max_dim)`,
            where `max_dim` corresponds to the output slice dimensions from the
            forward pass. `(...)` represents optional batch dimensions.

        Returns
        -------
            The 3D Volume. The shape will be `(..., Z, Y, X)`,
            matching the original `input_shape` of the forward operation.
        """
        match (self.matrix, self.matrix_adjoint):
            # selection based on the optimize_for setting
            case (None, None):
                raise RuntimeError('Either matrix or matrix adjoint must be set')
            case (matrix, None) if matrix is not None:
                matrix_adjoint = matrix.H
            case (None, matrix_adjoint) if matrix_adjoint is not None:
                matrix = matrix_adjoint.H
            case (matrix, matrix_adjoint):
                ...

        # For the (unusual case) of batched volumes, we will apply for each element in series
        n_batchdim = len(self._range_shape) - 3
        # x_domainbatch_range has all volume batch dimensions moved to the front
        x_domainbatch_range = x.moveaxis(tuple(range(n_batchdim, x.ndim - 3)), tuple(range(x.ndim - 3 - n_batchdim)))
        # x_flatdomainbatch_flatrange is 2D with shape
        # (all batch dimensions of the volume flattened, all range dimensions flattened)
        x_domainbatch_flatrange = torch.atleast_2d(x_domainbatch_range.flatten(start_dim=-len(self._range_shape)))
        x_flatdomainbatch_flatrange = x_domainbatch_flatrange.flatten(end_dim=-2)
        # y_flatdomainbatch has shape (all batch dimensions of the volume flattened, *range dimensions)
        y_flatdomainbatch = torch.stack(
            [
                _MatrixMultiplication.apply(x, matrix_adjoint, matrix).reshape(self._domain_shape)
                for x in x_flatdomainbatch_flatrange
            ]
        )
        y = y_flatdomainbatch.reshape(*x.shape[n_batchdim:-3], *y_flatdomainbatch.shape[1:])
        return (y,)

    @staticmethod
    def join_matrices(matrices: Sequence[Tensor]) -> Tensor:
        """Join multiple sparse matrices into a block diagonal matrix.

        Parameters
        ----------
        matrices
            List of sparse matrices to join by stacking them as a block diagonal matrix
        """
        values = []
        target = []
        source = []
        for i, m in enumerate(matrices):
            if not m.shape == matrices[0].shape:
                raise ValueError('all matrices should have the same shape')
            c = m.coalesce()  # we want unique indices
            (ctarget, csource) = c.indices()
            values.append(c.values())
            source.append(csource)
            ctarget = ctarget + i * m.shape[0]
            target.append(ctarget)

        with warnings.catch_warnings():
            # beta status in pytorch causes a warning to be printed
            warnings.filterwarnings('ignore', category=UserWarning, message='Sparse')
            matrix = torch.sparse_coo_tensor(
                indices=torch.stack([torch.cat(target), torch.cat(source)]),
                values=torch.cat(values),
                dtype=torch.float32,
                size=(len(matrices) * m.shape[0], m.shape[1]),
            )
        return matrix

    @staticmethod
    def projection_matrix(
        input_shape: SpatialDimension[int],
        output_shape: SpatialDimension[int],
        rotation: Rotation,
        offset: Tensor,
        w: int,
        slice_function: TensorFunction,
        rotation_center: Tensor | None = None,
    ) -> Tensor:
        """Create a sparse matrix that represents the projection of a volume onto a plane.

        Outside the volume values are approximately zero padded

        Parameters
        ----------
        input_shape
            Shape of the volume to sample from
        output_shape
            Shape of the resulting plane, 2D. Only the x and y values are used.
        rotation
            Rotation that describes the orientation of the plane
        offset: Tensor
            Shift of the plane from the center of the volume in the rotated coordinate system
            in units of the 3D volume, order `z, y, x`
        w: int
            Factor that determines the number of pixels that are considered in the projection along
            the slice profile direction.
        slice_function
            Function that describes the slice profile. See `mrpro.utils.slice_profiles` for examples.
        rotation_center
            Center of rotation, if `None` the center of the volume is used,
            i.e. for 4 pixels `[0 1 2 3]` it is between 1 and 2

        Returns
        -------
        torch.sparse_coo_matrix
            Sparse matrix that represents the projection of the volume onto the plane
        """
        x, y = output_shape.x, output_shape.y

        start_x, start_y = (
            (input_shape.x - x) // 2,
            (input_shape.y - y) // 2,
        )
        pixel_coord_y_x_zyx = torch.stack(
            [
                (input_shape.z / 2 - 0.5) * torch.ones(y, x),  # z coordinates
                *torch.meshgrid(
                    torch.arange(start_y, start_y + y),  # y coordinates
                    torch.arange(start_x, start_x + x),  # x coordinates
                    indexing='ij',
                ),
            ],
            dim=-1,
        )  # coordinates of the 2d output pixels in the coordinate system of the input volume, so shape (y,x,3)
        if offset is not None:
            pixel_coord_y_x_zyx = pixel_coord_y_x_zyx + offset
        if rotation_center is None:
            # default rotation center is the center of the volume, i.e. for 4 pixels
            # 0 1 2 3 it is between 0 and 1
            rotation_center = torch.tensor([input_shape.z / 2 - 0.5, input_shape.y / 2 - 0.5, input_shape.x / 2 - 0.5])
        pixel_rotated_y_x_zyx = rotation(pixel_coord_y_x_zyx - rotation_center) + rotation_center

        # We cast a ray from the pixel normal to the plane in both directions
        # points in the original volume further away then w will not be considered
        ray = rotation(
            torch.stack(
                [
                    torch.arange(-w, w + 1),  # z
                    torch.zeros(2 * w + 1),  # y
                    torch.zeros(2 * w + 1),  # x
                ],
                dim=-1,
            )
        )
        # In all possible directions for each point along the line we consider the eight neighboring points
        # by adding all possible combinations of 0 and 1 to the point and flooring
        offsets = torch.tensor(list(itertools.product([0, 1], repeat=3)))
        # all points that influence a pixel
        # x,y,8-neighbors,(2*w+1)-raylength,3-dimensions input_shape.xinput_shape.yinput_shape.z)
        points_influencing_pixel = (
            einops.rearrange(pixel_rotated_y_x_zyx, '   y x zyxdim -> y x 1          1   zyxdim')
            + einops.rearrange(ray, '                   ray zyxdim -> 1 1 1          ray zyxdim')
            + einops.rearrange(offsets, '        neighbors zyxdim -> 1 1 neighbors 1   zyxdim')
        ).floor()  # y x neighbors ray zyx
        # directional distance in source volume coordinate system
        distance = pixel_rotated_y_x_zyx[:, :, None, None, :] - points_influencing_pixel
        # Inverse rotation projects this back to the original coordinate system, i.e
        # Distance in z is distance along the line, i.e. the slice profile weighted direction
        # Distance in x and y is the distance of a pixel to the ray and linear interpolation
        # is used to weight the distance
        distance_z, distance_y, distance_x = rotation(distance, inverse=True).unbind(-1)
        weight_yx = (1 - distance_y.abs()).clamp_min(0) * (1 - distance_x.abs()).clamp_min(0)
        weight_z = slice_function(distance_z)
        weight = (weight_yx * weight_z).reshape(y * x, -1)

        source = einops.rearrange(
            points_influencing_pixel,
            'y x neighbors raylength zyxdim -> (y x) (neighbors raylength) zyxdim',
        ).int()

        # mask of only potential source points inside the source volume
        mask = (
            (source[..., 0] < input_shape.z)
            & (source[..., 0] >= 0)
            & (source[..., 1] < input_shape.y)
            & (source[..., 1] >= 0)
            & (source[..., 2] < input_shape.x)
            & (source[..., 2] >= 0)
        )

        # We need this at the edge of the volume to approximate zero padding
        fraction_in_view = (mask * (weight > 0)).sum(-1) / (weight > 0).sum(-1)

        source_index = ravel_multi_index(source[mask].unbind(-1), (input_shape.z, input_shape.y, input_shape.x))

        target_index = torch.repeat_interleave(torch.arange(y * x), mask.sum(-1))

        with warnings.catch_warnings():
            # beta status in pytorch causes a warning to be printed
            warnings.filterwarnings('ignore', category=UserWarning, message='Sparse')

            matrix = torch.sparse_coo_tensor(
                indices=torch.stack((target_index, source_index)),
                values=weight.reshape(y * x, -1)[mask],
                size=(y * x, input_shape.z * input_shape.y * input_shape.x),
                dtype=torch.float32,
            ).coalesce()

            # To avoid giving more weight to points that are duplicated in our ray
            # logic and got summed in the coalesce operation, we normalize by the number
            # of duplicates. This is equivalent to the sum of the weights of the duplicates.
            # Count duplicates...

            ones = torch.ones_like(source_index).float()
            ones = torch.sparse_coo_tensor(
                indices=torch.stack((target_index, source_index)),
                values=ones,
                size=(y * x, input_shape.z * input_shape.y * input_shape.x),
                dtype=torch.float32,
            )
            # Coalesce sums the values of duplicate indices
            ones = ones.coalesce()

        # .. and normalize by the number of duplicates
        matrix.values()[:] /= ones.values()

        # Normalize for slice profile, so that the sum of the weights is 1
        # independent of the number of points that are considered.
        # Within the volume, the column sum should be 1,
        # at the edge of the volume, the column sum should be the fraction of the slice
        # that is in view to approximate zero padding
        norm = fraction_in_view / (matrix.sum(1).to_dense() + 1e-6)
        matrix *= norm[:, None]
        return matrix



"""MR raw data / k-space data class."""

import copy
import dataclasses
import datetime
import warnings
from collections.abc import Callable, Mapping, Sequence
from importlib.metadata import version
from types import EllipsisType
from typing import Literal, cast

import h5py
import ismrmrd
import numpy as np
import torch
from einops import rearrange
from typing_extensions import Self, TypeVar

from mrpro.data.acq_filters import has_n_coils, is_image_acquisition
from mrpro.data.AcqInfo import (
    AcqInfo,
    convert_time_stamp_from_osi2,
    convert_time_stamp_from_siemens,
    convert_time_stamp_to_osi2,
    convert_time_stamp_to_siemens,
    write_acqinfo_to_ismrmrd_acquisition_,
)
from mrpro.data.Dataclass import Dataclass
from mrpro.data.EncodingLimits import EncodingLimits, Limits
from mrpro.data.enums import AcqFlags
from mrpro.data.KHeader import KHeader
from mrpro.data.KTrajectory import KTrajectory
from mrpro.data.Rotation import Rotation
from mrpro.data.traj_calculators.KTrajectoryCalculator import KTrajectoryCalculator
from mrpro.data.traj_calculators.KTrajectoryIsmrmrd import KTrajectoryIsmrmrd
from mrpro.utils.summarize import summarize_object
from mrpro.utils.typing import FileOrPath

RotationOrTensor = TypeVar('RotationOrTensor', bound=torch.Tensor | Rotation)

KDIM_SORT_LABELS = (
    'k1',
    'k2',
    'average',
    'slice',
    'contrast',
    'phase',
    'repetition',
    'set',
    'user0',
    'user1',
    'user2',
    'user3',
    'user4',
    'user7',
)

OTHER_LABELS = (
    'average',
    'slice',
    'contrast',
    'phase',
    'repetition',
    'set',
    'user0',
    'user1',
    'user2',
    'user3',
    'user4',
    'user7',
)

T = TypeVar('T', Rotation, torch.Tensor, Rotation | torch.Tensor)


class KData(Dataclass):
    """MR raw data / k-space data class."""

    header: KHeader
    """Header information for k-space data"""

    data: torch.Tensor
    """K-space data. Shape `(*other coils k2 k1 k0)`"""

    traj: KTrajectory
    """K-space trajectory along kz, ky and kx. Shape `(*other 1 k2 k1 k0)`"""

    @classmethod
    def from_file(
        cls,
        filename: FileOrPath,
        trajectory: KTrajectoryCalculator | KTrajectory | KTrajectoryIsmrmrd,
        header_overwrites: Mapping[str, object] | None = None,
        dataset_idx: int = -1,
        acquisition_filter_criterion: Callable = is_image_acquisition,
    ) -> Self:
        """Load k-space data from an ISMRMRD file.

        Parameters
        ----------
        filename
            path to the ISMRMRD file or file-like object
        trajectory
            KTrajectoryCalculator to calculate the k-space trajectory or an already calculated KTrajectory
            If a KTrajectory is given, the shape should be `(acquisisions 1 1 k0)` in the same order as the acquisitions
            in the ISMRMRD file.
        header_overwrites
            dictionary of key-value pairs to overwrite the header
        dataset_idx
            index of the ISMRMRD dataset to load (converter creates dataset, dataset_1, ...)
        acquisition_filter_criterion
            function which returns True if an acquisition should be included in KData
        """
        # Can raise FileNotFoundError
        with ismrmrd.File(filename, 'r') as file:
            dataset = file[list(file.keys())[dataset_idx]]
            ismrmrd_header = dataset.header
            acquisitions = dataset.acquisitions[:]
            try:
                mtime: int = h5py.h5g.get_objinfo(dataset['data']._contents.id).mtime
            except AttributeError:
                mtime = 0
            modification_time = datetime.datetime.fromtimestamp(mtime)

        acquisitions = [acq for acq in acquisitions if acquisition_filter_criterion(acq)]

        # we need the same number of receiver coils for all acquisitions
        n_coils_available = {acq.data.shape[0] for acq in acquisitions}
        if len(n_coils_available) > 1:
            if (
                ismrmrd_header.acquisitionSystemInformation is not None
                and ismrmrd_header.acquisitionSystemInformation.receiverChannels is not None
            ):
                n_coils = int(ismrmrd_header.acquisitionSystemInformation.receiverChannels)
            else:
                # most likely, highest number of elements are the coils used for imaging
                n_coils = int(max(n_coils_available))

            warnings.warn(
                f'Acquisitions with different number {n_coils_available} of receiver coil elements detected. '
                f'Data with {n_coils} receiver coil elements will be used.',
                stacklevel=1,
            )
            acquisitions = [acq for acq in acquisitions if has_n_coils(n_coils, acq)]

        if not acquisitions:
            raise ValueError('No acquisitions meeting the given filter criteria were found.')

        if ismrmrd_header.acquisitionSystemInformation is not None and isinstance(
            ismrmrd_header.acquisitionSystemInformation.systemVendor, str
        ):
            match ismrmrd_header.acquisitionSystemInformation.systemVendor.lower():
                case 'siemens':
                    convert_time_stamp = convert_time_stamp_from_siemens  # 2.5ms time steps
                case 'osi2':
                    convert_time_stamp = convert_time_stamp_from_osi2  # 1ms time steps
                case str(vendor):
                    warnings.warn(
                        f'Unknown vendor {vendor}. '
                        'Assuming Siemens time stamp format. If this is wrong, consider opening an Issue.',
                        stacklevel=1,
                    )
                    convert_time_stamp = convert_time_stamp_from_siemens  # 2.5ms time steps
        else:
            warnings.warn('No vendor information found. Assuming Siemens time stamp format.', stacklevel=1)
            convert_time_stamp = convert_time_stamp_from_siemens

        acq_info, (k0_center, n_k0_tensor, discard_pre, discard_post) = AcqInfo.from_ismrmrd_acquisitions(
            acquisitions,
            additional_fields=('center_sample', 'number_of_samples', 'discard_pre', 'discard_post'),
            convert_time_stamp=convert_time_stamp,
        )
        discard_pre_1d = rearrange(discard_pre, 'n_readouts 1 1 1 1 -> n_readouts')
        discard_post_1d = rearrange(discard_post, 'n_readouts 1 1 1 1 -> n_readouts')

        if len(torch.unique(acq_info.idx.user5)) > 1:
            warnings.warn(
                'The Siemens to ismrmrd converter currently (ab)uses '
                'the user 5 indices for storing the kspace center line number.\n'
                'User 5 indices will be ignored',
                stacklevel=1,
            )

        if len(torch.unique(acq_info.idx.user6)) > 1:
            warnings.warn(
                'The Siemens to ismrmrd converter currently (ab)uses '
                'the user 6 indices for storing the kspace center partition number.\n'
                'User 6 indices will be ignored',
                stacklevel=1,
            )

        shapes = (
            torch.as_tensor([acq.data.shape[-1] for acq in acquisitions]) - discard_pre_1d - discard_post_1d
        ).unique()
        if len(shapes) > 1:
            warnings.warn(
                f'Acquisitions have different shape. Got {list(shapes)}. '
                f'Keeping only acquisistions with {shapes[-1]} data samples. Note: discard_pre and discard_post '
                f'{"have been applied. " if discard_pre_1d.any() or discard_post_1d.any() else "were empty. "}'
                'Please open an issue of you need to handle this kind of data.',
                stacklevel=1,
            )
        data = torch.stack(
            [
                torch.as_tensor(acq.data[..., pre : acq.data.shape[-1] - post], dtype=torch.complex64)
                for acq, pre, post in zip(acquisitions, discard_pre_1d, discard_post_1d, strict=True)
                if acq.data.shape[-1] - pre - post == shapes[-1]
            ]
        )
        n_k0_tensor = n_k0_tensor - discard_pre - discard_post
        data = rearrange(data, 'acquisitions coils k0 -> acquisitions coils 1 1 k0')

        # Raises ValueError if required fields are missing in the header
        header = KHeader.from_ismrmrd(
            ismrmrd_header,
            acq_info,
            defaults={
                'datetime': modification_time,  # use the modification time of the dataset as fallback
                'trajectory': trajectory,
            },
            overwrite=header_overwrites,
        )
        if header.vendor.lower() == 'siemens':
            # Siemens assumes a fft to go from k-space to image space
            data = data.conj_physical()

        # Calculate trajectory and check if it matches the kdata shape
        match trajectory:
            case KTrajectoryIsmrmrd():
                trajectory_ = trajectory(acquisitions, encoding_matrix=header.encoding_matrix)
            case KTrajectoryCalculator():
                reversed_readout_mask = (header.acq_info.flags[..., 0] & AcqFlags.ACQ_IS_REVERSE.value).bool()
                n_k0_unique = torch.unique(n_k0_tensor)
                if len(n_k0_unique) > 1:
                    raise ValueError(
                        'Trajectory can only be calculated for constant number of readout samples.\n'
                        f'Got unique values {list(n_k0_unique)}'
                    )
                encoding_limits = EncodingLimits.from_ismrmrd_header(ismrmrd_header)
                trajectory_ = trajectory(
                    n_k0=int(n_k0_unique[0]),
                    k0_center=k0_center,
                    k1_idx=header.acq_info.idx.k1,
                    k1_center=encoding_limits.k1.center,
                    k2_idx=header.acq_info.idx.k2,
                    k2_center=encoding_limits.k2.center,
                    reversed_readout_mask=reversed_readout_mask,
                    encoding_matrix=header.encoding_matrix,
                )
            case KTrajectory():
                try:
                    torch.broadcast_shapes(trajectory.shape, (data.shape[0], *data.shape[-3:]))
                except RuntimeError:
                    raise ValueError(
                        f'Trajectory shape {trajectory.shape} does not match data shape {data.shape}.'
                    ) from None
                trajectory_ = trajectory
            case _:
                raise TypeError(
                    'ktrajectory must be KTrajectoryIsmrmrd, KTrajectory or KTrajectoryCalculator'
                    f'not {type(trajectory)}',
                )

        kdata = cls(header, data, trajectory_)
        kdata = kdata.reshape_by_idx()
        return kdata

    def reshape_by_idx(self) -> Self:
        """Sort and reshape according to the acquisistion indices.

        Reshapes the data to ("all other labels", coils, k2, k1, k0).
        Within "all other labels", the order is determined by `KDIM_SORT_LABELS`.
        """
        shape = self.shape

        def expand(x: T) -> T:
            x = x.expand(*shape[:-4], -1, shape[-3], shape[-2], -1)
            return x

        # First, determine if we can split into k2 and k1 and how large these should be
        acq_indices_other = torch.stack(
            [expand(getattr(self.header.acq_info.idx, label)).ravel() for label in OTHER_LABELS],
            dim=0,
        )
        _, n_acqs_per_other = torch.unique(acq_indices_other, dim=1, return_counts=True)
        # unique counts of acquisitions for each combination of the label values in "other"
        n_acqs_per_other = torch.unique(n_acqs_per_other)

        acq_indices_other_k2 = torch.cat(
            (acq_indices_other, expand(self.header.acq_info.idx.k2).ravel().unsqueeze(0)), dim=0
        )
        _, n_acqs_per_other_and_k2 = torch.unique(acq_indices_other_k2, dim=1, return_counts=True)
        # unique counts of acquisitions for each combination of other **and k2**
        n_acqs_per_other_and_k2 = torch.unique(n_acqs_per_other_and_k2)

        if len(n_acqs_per_other_and_k2) == 1:
            # This is the most common case:
            # All other and k2 combinations have the same number of aquisistions which we can use as k1
            # to reshape to (other, coils, k2, k1, k0)
            n_k1 = n_acqs_per_other_and_k2[0]
            n_k2 = n_acqs_per_other[0] // n_k1
        elif len(n_acqs_per_other) == 1:
            # We cannot split the data along phase encoding steps into k2 and k1, as there are different numbers of k1.
            # But all "other labels" combinations have the same number of aquisisitions,
            # so we can reshape to (other, coils, k, 1, k0)
            n_k1 = 1
            n_k2 = n_acqs_per_other[0]  # total number of k encoding steps
        else:
            # For different other combinations we have different numbers of aquisistions,
            # so we can only reshape to (acquisitions, coils, 1, 1, k0)
            # This might be an user error.
            warnings.warn(
                f'There are different numbers of acquisistions in'
                'different combinations of labels {"/".join(OTHER_LABELS)}: \n'
                f'Found {n_acqs_per_other.tolist()}. \n'
                'The data will be reshaped to (acquisitions, coils, 1, 1, k0). \n'
                'This needs to be adjusted be reshaping for a successful reconstruction. \n'
                'If unintenional, this might be caused by wrong labels in the ismrmrd file or a wrong flag filter. \n'
                'To fix, it might be necessary to subset the data such that it can be reshaped to '
                '(other, coils, k2, k1, k0), or indexes needs to be fixed. \n'
                'After fixing the issue, call reshape_by_idx and consider recalculating the trajectory.',
                stacklevel=1,
            )
            n_k1 = 1
            n_k2 = 1

        # Second, determine the sorting order
        acq_indices = np.stack(
            [expand(getattr(self.header.acq_info.idx, label)).ravel() for label in KDIM_SORT_LABELS],
            axis=0,
        )
        sort_idx = torch.as_tensor(np.lexsort(acq_indices))  # torch has no lexsort as of pytorch 2.6 (March 2025)

        # Finally, reshape and sort the tensors in acqinfo and acqinfo.idx, and kdata.
        def sort(x: T) -> T:
            flat = cast(T, rearrange(expand(x), '... coils k2 k1 k0 -> (... k2 k1) coils k0'))
            return cast(
                T, rearrange(flat[sort_idx], '(other k2 k1) coils k0 -> other coils k2 k1 k0', k1=n_k1, k2=n_k2)
            )

        header = self.header.apply(lambda field: sort(field) if isinstance(field, torch.Tensor | Rotation) else field)
        data = sort(self.data)
        kz, ky, kx = (sort(t) for t in (self.traj.kz, self.traj.ky, self.traj.kx))
        traj = KTrajectory(kz, ky, kx, self.traj.grid_detection_tolerance, self.traj.repeat_detection_tolerance)
        return type(self)(header=header._reduce_repeats_(), data=data, traj=traj._reduce_repeats_())

    def to_file(self, filename: FileOrPath) -> None:
        """Save KData as ISMRMRD dataset to file.

        Parameters
        ----------
        filename
            path to the ISMRMRD file
        """
        ismrmrd_version = int(version('ismrmrd').split('.')[0])
        header = self.header.to_ismrmrd()
        header.acquisitionSystemInformation.receiverChannels = self.data.shape[-4]

        # Calculate the encoding limits as min/max of the acquisition indices
        def limits_from_acq_idx(acq_idx_tensor: torch.Tensor) -> Limits:
            return Limits(int(acq_idx_tensor.min().item()), int(acq_idx_tensor.max().item()), 0)

        encoding_limits = EncodingLimits(
            **{
                field.name: limits_from_acq_idx(getattr(self.header.acq_info.idx, field.name))
                for field in dataclasses.fields(self.header.acq_info.idx)
            }
        )

        # For the k-space center of k1 and k2 we can only make an educated guess on where it is:
        # k-space point closest to 0
        center_idx = self.traj.as_tensor(-1).abs().sum(dim=-1).argmin()
        encoding_limits.k1.center = int(self.header.acq_info.idx.k1.broadcast_to(self.traj.shape).flatten()[center_idx])
        encoding_limits.k2.center = int(self.header.acq_info.idx.k2.broadcast_to(self.traj.shape).flatten()[center_idx])
        header.encoding[0].encodingLimits = encoding_limits.to_ismrmrd_encoding_limits_type()

        # Vendors use different units for time stamps
        if self.header.vendor.lower() == 'osi2':
            convert_time_stamp = convert_time_stamp_to_osi2  # 1ms time steps
        else:
            convert_time_stamp = convert_time_stamp_to_siemens  # 2.5ms time steps

        with ismrmrd.Dataset(filename, 'dataset', create_if_needed=True) as dataset:
            dataset.write_xml_header(header.toXML('utf-8'))

            flattened = self.rearrange('... coils k2 k1 k0 -> (... k2 k1) coils 1 1 k0')
            for scan_counter, acq in enumerate(flattened):
                ismrmrd_acq = ismrmrd.Acquisition()
                ismrmrd_acq.resize(
                    number_of_samples=acq.shape[-1], active_channels=acq.shape[-4], trajectory_dimensions=3
                )
                ismrmrd_acq.available_channels = acq.shape[-4]
                ismrmrd_acq.version = ismrmrd_version
                ismrmrd_acq.scan_counter = 0
                write_acqinfo_to_ismrmrd_acquisition_(acq.header.acq_info, ismrmrd_acq, convert_time_stamp)
                ismrmrd_acq.traj[:] = acq.traj.as_tensor(-1).squeeze().cpu().numpy()[:, ::-1]  # zyx -> xyz
                ismrmrd_acq.center_sample = np.argmin(np.abs(ismrmrd_acq.traj[:, 0]))
                acq_data = acq.data.squeeze().cpu().numpy()
                # Siemens assumes a fft to go from k-space to image space
                ismrmrd_acq.data[:] = acq_data.conj() if self.header.vendor.lower() == 'siemens' else acq_data
                ismrmrd_acq.scan_counter = scan_counter
                dataset.append_acquisition(ismrmrd_acq)

    def __repr__(self):
        """Representation method for KData class."""
        traj_info = '\n   '.join(repr(self.traj).splitlines())
        header_info = '\n   '.join(repr(self.header).splitlines())
        representation = '\n'.join(
            [
                super().__repr__().splitlines()[0],
                f'  data: {summarize_object(self.data)}',
                f'  traj: {traj_info}',
                f'  header:  {header_info}',
            ]
        )
        return representation

    def compress_coils(
        self: Self,
        n_compressed_coils: int,
        batch_dims: None | Sequence[int] = None,
        joint_dims: Sequence[int] | EllipsisType = ...,
    ) -> Self:
        """Reduce the number of coils based on a PCA compression.

        A PCA is carried out along the coil dimension and the n_compressed_coils virtual coil elements are selected. For
        more information on coil compression please see [BUE2007]_, [DON2008]_ and [HUA2008]_.

        Returns a copy of the data.

        Parameters
        ----------
        kdata
            K-space data
        n_compressed_coils
            Number of compressed coils
        batch_dims
            Dimensions which are treated as batched, i.e. separate coil compression matrizes (e.g. different slices).
            Default is to do one coil compression matrix for the entire k-space data. Only batch_dim or joint_dim can
            be defined. If batch_dims is not None then joint_dims has to be ...
        joint_dims
            Dimensions which are combined to calculate single coil compression matrix (e.g. k0, k1, contrast). Default
            is that all dimensions (except for the coil dimension) are joint_dims. Only batch_dim or joint_dim can
            be defined. If joint_dims is not ... batch_dims has to be None

        Returns
        -------
            Copy of K-space data with compressed coils.

        Raises
        ------
        `ValueError`
            If both batch_dims and joint_dims are defined.
        `ValuerError`
            If coil dimension is part of joint_dims or batch_dims.

        References
        ----------
        .. [BUE2007] Buehrer M, Pruessmann KP, Boesiger P, Kozerke S (2007) Array compression for MRI with large coil
           arrays. MRM 57. https://doi.org/10.1002/mrm.21237
        .. [DON2008] Doneva M, Boernert P (2008) Automatic coil selection for channel reduction in SENSE-based parallel
           imaging. MAGMA 21. https://doi.org/10.1007/s10334-008-0110-x
        .. [HUA2008] Huang F, Vijayakumar S, Li Y, Hertel S, Duensing GR (2008) A software channel compression
           technique for faster reconstruction with many channels. MRM 26. https://doi.org/10.1016/j.mri.2007.04.010

        """
        from mrpro.operators import PCACompressionOp

        coil_dim = -4 % self.data.ndim

        if n_compressed_coils > (n_current_coils := self.data.shape[coil_dim]):
            raise ValueError(
                f'Number of compressed coils ({n_compressed_coils}) cannot be greater '
                f'than the number of current coils ({n_current_coils}).'
            )

        if batch_dims is not None and joint_dims is not Ellipsis:
            raise ValueError('Either batch_dims or joint_dims can be defined not both.')

        if joint_dims is not Ellipsis:
            joint_dims_normalized = [i % self.data.ndim for i in joint_dims]
            if coil_dim in joint_dims_normalized:
                raise ValueError('Coil dimension must not be in joint_dims')
            batch_dims_normalized = [
                d for d in range(self.data.ndim) if d not in joint_dims_normalized and d is not coil_dim
            ]
        else:
            batch_dims_normalized = [] if batch_dims is None else [i % self.data.ndim for i in batch_dims]
            if coil_dim in batch_dims_normalized:
                raise ValueError('Coil dimension must not be in batch_dims')

        # reshape to (*batch dimension, -1, coils)
        permute_order = (
            *batch_dims_normalized,
            *[i for i in range(self.data.ndim) if i != coil_dim and i not in batch_dims_normalized],
            coil_dim,
        )
        kdata_permuted = self.data.permute(permute_order)
        kdata_flattened = kdata_permuted.flatten(
            start_dim=len(batch_dims_normalized), end_dim=-2
        )  # keep separate dimensions and coil

        pca_compression_op = PCACompressionOp(data=kdata_flattened, n_components=n_compressed_coils)
        (kdata_coil_compressed_flattened,) = pca_compression_op(kdata_flattened)
        del kdata_flattened
        # reshape to original dimensions and undo permutation
        kdata_coil_compressed = torch.reshape(
            kdata_coil_compressed_flattened, [*kdata_permuted.shape[:-1], n_compressed_coils]
        ).permute(*np.argsort(permute_order))

        return type(self)(self.header.clone(), kdata_coil_compressed, self.traj.clone())

    def remove_readout_os(self: Self) -> Self:
        """Remove any oversampling along the readout direction.

        Removes oversampling along the readout direction by cropping the data
        to the size of the reconstruction matrix in image space [GAD]_.

        Returns a copy of the data.

        Parameters
        ----------
        kdata
            K-space data

        Returns
        -------
            Copy of K-space data with oversampling removed.

        Raises
        ------
        `ValueError`
            If the recon matrix along x is larger than the encoding matrix along x.

        References
        ----------
        .. [GAD] Gadgetron https://github.com/gadgetron/gadgetron-python
        """
        from mrpro.operators.FastFourierOp import FastFourierOp

        # Ratio of k0/x between encoded and recon space
        x_ratio = self.header.recon_matrix.x / self.header.encoding_matrix.x
        if x_ratio == 1:
            # If the encoded and recon space is the same we don't have to do anything
            return self
        elif x_ratio > 1:
            raise ValueError('Recon matrix along x should be equal or larger than encoding matrix along x.')

        # Starting and end point of image after removing oversampling
        start_cropped_readout = (self.header.encoding_matrix.x - self.header.recon_matrix.x) // 2
        end_cropped_readout = start_cropped_readout + self.header.recon_matrix.x

        def crop_readout(data_to_crop: torch.Tensor) -> torch.Tensor:
            # returns a cropped copy
            return data_to_crop[..., start_cropped_readout:end_cropped_readout].clone()

        # Transform to image space along readout, crop to reconstruction matrix size and transform back
        fourier_k0_op = FastFourierOp(dim=(-1,))
        (cropped_data,) = fourier_k0_op(crop_readout(*fourier_k0_op.H(self.data)))

        # Adapt trajectory
        ks = [self.traj.kz, self.traj.ky, self.traj.kx]
        # only cropped ks that are not broadcasted/singleton along k0
        cropped_ks = [crop_readout(k) if k.shape[-1] > 1 else k.clone() for k in ks]
        cropped_traj = KTrajectory(cropped_ks[0], cropped_ks[1], cropped_ks[2])

        # Adapt header parameters
        header = copy.deepcopy(self.header)
        header.encoding_matrix.x = cropped_data.shape[-1]
        header.encoding_fov.x = self.header.recon_fov.x

        return type(self)(header, cropped_data, cropped_traj)

    def select_other_subset(
        self: Self,
        subset_idx: torch.Tensor,
        subset_label: Literal['average', 'slice', 'contrast', 'phase', 'repetition', 'set'],
    ) -> Self:
        """Select a subset from the other dimension of KData.

        .. warning::
            This function will be deprecated in the future.

        Parameters
        ----------
        kdata
            K-space data `(other coils k2 k1 k0)`
        subset_idx
            Index which elements of the other subset to use, e.g. phase 0,1,2 and 5
        subset_label
            Name of the other label, e.g. phase

        Returns
        -------
            K-space data `(other_subset coils k2 k1 k0)`

        Raises
        ------
        `ValueError`
            If the subset indices are not available in the data
        """
        # Flatten multi-dimensional other
        n_other = self.data.shape[:-4]  # Assume that data is not broadcasted along other
        header = self.header.apply(
            lambda field: rearrange(
                field.expand(*n_other, *field.shape[-4:]), '... coils k2 k1 k0->(...) coils k2 k1 k0'
            )
            if isinstance(field, torch.Tensor | Rotation)
            else field
        )
        traj = self.traj.as_tensor()
        traj = torch.broadcast_to(traj, (traj.shape[0], *self.data.shape[:-4], *traj.shape[-4:]))  # broadcast "other"
        traj = traj.flatten(start_dim=1, end_dim=-5)  # flatten "other" dimensions
        data = self.data.flatten(end_dim=-5)

        # Find elements in the subset
        label_idx = getattr(header.acq_info.idx, subset_label)
        if not all(el in torch.unique(label_idx) for el in subset_idx):
            raise ValueError('Subset indices are outside of the available index range')
        other_idx = torch.cat([torch.where(idx == label_idx[:, 0, 0, 0, 0])[0] for idx in subset_idx], dim=0)

        # Select subset
        header.acq_info.apply_(lambda field: field[other_idx] if isinstance(field, torch.Tensor | Rotation) else field)
        data = data[other_idx]
        traj = traj[:, other_idx]
        return type(self)(header, data, type(self.traj).from_tensor(traj))



"""Total Variation (TV) Denoising using PDHG."""

from __future__ import annotations

from collections.abc import Sequence
from typing import overload

import torch

from mrpro.algorithms.optimizers.pdhg import pdhg
from mrpro.data.IData import IData
from mrpro.operators import LinearOperatorMatrix, ProximableFunctionalSeparableSum
from mrpro.operators.FiniteDifferenceOp import FiniteDifferenceOp
from mrpro.operators.functionals import L1NormViewAsReal, L2NormSquared
from mrpro.operators.IdentityOp import IdentityOp
from mrpro.utils import normalize_index, unsqueeze_right


@overload
def total_variation_denoising(
    idata: IData,
    regularization_dim: Sequence[int],
    regularization_weight: float | Sequence[float] | Sequence[torch.Tensor],
    initial_image: torch.Tensor | None = None,
    max_iterations: int = 100,
    tolerance: float = 0,
) -> IData: ...


@overload
def total_variation_denoising(
    idata: torch.Tensor,
    regularization_dim: Sequence[int],
    regularization_weight: float | Sequence[float] | Sequence[torch.Tensor],
    initial_image: torch.Tensor | None = None,
    max_iterations: int = 100,
    tolerance: float = 0,
) -> torch.Tensor: ...


def total_variation_denoising(
    idata: IData | torch.Tensor,
    regularization_dim: Sequence[int],
    regularization_weight: float | Sequence[float] | Sequence[torch.Tensor],
    initial_image: torch.Tensor | None = None,
    max_iterations: int = 100,
    tolerance: float = 0,
) -> IData | torch.Tensor:
    r"""Apply total variation denoising.

    This algorithm solves the problem :math:`min_x \frac{1}{2}||x - y||_2^2 + \\sum_i l_i ||\nabla_i x||_1`
    by using the PDHG-algorithm. :math:`y` is the given noisy image, :math:`l_i` are the strengths of the regularization
    along the different dimensions and :math:`\nabla_i` is the finite difference operator applied to :math:`x` along
    different dimensions :math:`i`.

    Parameters
    ----------
    idata
        input image
    regularization_dim
        Dimensions along which the total variation reguarization is applied (:math:`i`).
    regularization_weight
        Strengths of the regularization (:math:`l_i`). If a single values is given, it is applied to all dimensions.
        If a sequence is given, it must have the same length as `regularization_dim`.
    initial_image
        Initial image. If `None` then the target image :math:`y` will be used.
    max_iterations
        Maximum number of PDHG iterations.
    tolerance
            Tolerance of PDHG for relative change of the primal solution; if zero, `max_iterations` of PDHG are run.

    Returns
    -------
        the denoised image.
    """
    regularization_weight_ = torch.as_tensor(
        regularization_weight
        if isinstance(regularization_weight, Sequence)
        else [regularization_weight] * len(regularization_dim)
    )
    img_tensor = idata if isinstance(idata, torch.Tensor) else idata.data

    if len(regularization_dim) != len(regularization_weight_):
        raise ValueError('Regularization dimensions and weights must have the same length')
    regularization_dim = tuple(normalize_index(img_tensor.ndim, idx) for idx in regularization_dim)
    if len(regularization_dim) != len(set(regularization_dim)):
        raise ValueError('Repeated values are not allowed in regularization_dim')

    l2_norm_squared = L2NormSquared(target=img_tensor)

    # TV regularization
    nabla_operator = FiniteDifferenceOp(dim=regularization_dim, mode='forward')
    l1_norm = L1NormViewAsReal(weight=unsqueeze_right(regularization_weight_, img_tensor.ndim))
    operator = LinearOperatorMatrix(((IdentityOp(),), (nabla_operator,)))

    initial_image = initial_image if initial_image is not None else img_tensor

    (img_tensor,) = pdhg(
        f=ProximableFunctionalSeparableSum(l2_norm_squared, l1_norm),
        g=None,
        operator=operator,
        initial_values=(initial_image,),
        max_iterations=max_iterations,
        tolerance=tolerance,
    )
    return img_tensor if isinstance(idata, torch.Tensor) else IData(img_tensor, idata.header)



"""Zero Operator."""

import torch

from mrpro.operators.LinearOperator import LinearOperator


class ZeroOp(LinearOperator):
    """A constant zero operator.

    This operator always returns zero when applied to a tensor.
    It is the neutral element of the addition of operators.
    """

    def __init__(self, keep_shape: bool = False):
        """Initialize the Zero Operator.

        Returns a constant zero, either as a scalar or as a tensor of the same shape as the input,
        depending on the value of keep_shape.
        Returning a scalar can save memory and computation time in some cases.

        Parameters
        ----------
        keep_shape
            If `True`, the shape of the input is kept.
            If `False`, the output is, regardless of the input shape, an integer scalar 0,
            which can broadcast to the input shape and dtype.
        """
        self.keep_shape = keep_shape
        super().__init__()

    def __call__(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply the zero operator to the input tensor.

        This operator returns a tensor of zeros. Depending on the `keep_shape`
        attribute set during initialization, the output will either be a
        tensor of zeros with the same shape as the input `x`, or a scalar zero.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
            A tensor of zeros. This will be `torch.zeros_like(x)`
            if `keep_shape` is `True`, or `torch.tensor(0)` if `keep_shape` is `False`.

        .. note::
            Prefer calling the instance of the ZeroOp operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        return super().__call__(x)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward of ZeroOp.

        .. note::
            Prefer calling the instance of the ZeroOp operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        if self.keep_shape:
            return (torch.zeros_like(x),)
        else:
            return (torch.tensor(0),)

    def adjoint(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply the adjoint of the zero operator.

        Since the zero operator is self-adjoint (mapping everything to zero),
        this method behaves identically to the forward operation. It returns
        a tensor of zeros, either with the same shape as input `x` or as a
        scalar zero, depending on `keep_shape`.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
            A tensor of zeros. This will be `torch.zeros_like(x)`
            if `keep_shape` is `True`, or `torch.tensor(0)` if `keep_shape` is `False`.

        """
        if self.keep_shape:
            return (torch.zeros_like(x),)
        else:
            return (torch.tensor(0),)

    @property
    def H(self) -> LinearOperator:  # noqa: N802
        """Adjoint of the Zero Operator."""
        return self



"""Classes for calculating k-space trajectories."""

from mrpro.data.traj_calculators.KTrajectoryCalculator import KTrajectoryCalculator
from mrpro.data.traj_calculators.KTrajectoryRpe import KTrajectoryRpe
from mrpro.data.traj_calculators.KTrajectorySunflowerGoldenRpe import KTrajectorySunflowerGoldenRpe
from mrpro.data.traj_calculators.KTrajectoryRadial2D import KTrajectoryRadial2D
from mrpro.data.traj_calculators.KTrajectoryIsmrmrd import KTrajectoryIsmrmrd
from mrpro.data.traj_calculators.KTrajectoryPulseq import KTrajectoryPulseq
from mrpro.data.traj_calculators.KTrajectoryCartesian import KTrajectoryCartesian
from mrpro.data.traj_calculators.KTrajectorySpiral2D import KTrajectorySpiral2D
__all__ = [
    "KTrajectoryCalculator",
    "KTrajectoryCartesian",
    "KTrajectoryIsmrmrd",
    "KTrajectoryPulseq",
    "KTrajectoryRadial2D",
    "KTrajectoryRpe",
    "KTrajectorySpiral2D",
    "KTrajectorySunflowerGoldenRpe"
]



"""Conjugate gradient operator."""

from collections.abc import Callable
from typing import TYPE_CHECKING, TypeVar

import torch
from torch.autograd.function import once_differentiable

from mrpro.algorithms.optimizers.cg import cg
from mrpro.operators.LinearOperator import LinearOperator
from mrpro.operators.LinearOperatorMatrix import LinearOperatorMatrix

LinearOperatorFactory = Callable[..., LinearOperator]
LinearOperatorMatrixFactory = Callable[..., LinearOperatorMatrix]
T = TypeVar('T', torch.Tensor, bool)


class ConjugateGradientCTX(torch.autograd.function.FunctionCtx):
    """Only used for type hinting."""

    saved_tensors: tuple[torch.Tensor, ...]
    needs_input_grad: tuple[bool, ...]
    len_solution: int
    tolerance: float
    max_iterations: int
    rhs_factory: Callable[..., tuple[torch.Tensor, ...]]
    operator_factory: Callable[..., LinearOperatorMatrix | LinearOperator]


class ConjugateGradientFunction(torch.autograd.Function):
    """Autograd function for the CG operator."""

    if TYPE_CHECKING:

        @classmethod
        def apply(
            cls,
            operator_factory: Callable[..., LinearOperatorMatrix | LinearOperator],
            rhs_factory: Callable[..., tuple[torch.Tensor, ...]],
            *inputs: torch.Tensor,
        ) -> tuple[torch.Tensor, ...]:
            """Apply the function. Required for mypy."""
            return super().apply(operator_factory, rhs_factory, *inputs)

    @staticmethod
    def forward(
        ctx: ConjugateGradientCTX,
        operator_factory: Callable[..., LinearOperatorMatrix | LinearOperator],
        rhs_factory: Callable[..., tuple[torch.Tensor, ...]],
        *inputs: torch.Tensor,
        initial_value: tuple[torch.Tensor, ...] | None = None,
        max_iterations: int = 10000,
        tolerance: float = 1e-7,
    ) -> tuple[torch.Tensor, ...]:
        """Forward pass of the conjugate gradient operator."""
        operator = operator_factory(*inputs)
        rhs = rhs_factory(*inputs)
        rhs_norm = sum((r.abs().square().sum() for r in rhs), torch.tensor(0.0)).sqrt().item()
        tol_ = tolerance * max(rhs_norm, 1e-6)  # clip in case rhs is 0
        if isinstance(operator, LinearOperator):
            if len(rhs) != 1:
                raise ValueError('LinearOperator requires a single right-hand side tensor.')
            if initial_value is not None and len(initial_value) != 1:
                raise ValueError('LinearOperator requires a single initial value tensor.')
            solution: tuple[torch.Tensor, ...] = cg(
                operator, rhs, initial_value=initial_value, tolerance=tol_, max_iterations=max_iterations
            )
        else:
            solution = cg(operator, rhs, initial_value=initial_value, tolerance=tol_, max_iterations=max_iterations)
        ctx.save_for_backward(*solution, *inputs)
        ctx.len_solution = len(solution)
        ctx.tolerance = tolerance
        ctx.max_iterations = max_iterations
        ctx.rhs_factory = rhs_factory
        ctx.operator_factory = operator_factory
        return solution

    @staticmethod
    @once_differentiable
    def backward(ctx: ConjugateGradientCTX, *grad_output: torch.Tensor) -> tuple[torch.Tensor | None, ...]:
        """Backward pass of the conjugate gradient operator."""
        solution, inputs = (
            ctx.saved_tensors[: ctx.len_solution],
            ctx.saved_tensors[ctx.len_solution :],
        )
        inputs = tuple(x.detach().clone().requires_grad_(x.requires_grad) for x in inputs)
        with torch.enable_grad():
            rhs = ctx.rhs_factory(*inputs)
            operator = ctx.operator_factory(*inputs)
        inputs_with_grad = tuple(x for x, need_grad in zip(inputs, ctx.needs_input_grad[2:], strict=True) if need_grad)
        if inputs_with_grad:
            rhs_norm = sum((r.abs().square().sum() for r in grad_output), torch.tensor(0.0)).sqrt().item()
            tol_ = ctx.tolerance * max(rhs_norm, 1e-6)  # clip in case rhs is 0
            with torch.no_grad():
                if isinstance(operator, LinearOperatorMatrix):
                    z = cg(operator.H, grad_output, tolerance=tol_, max_iterations=ctx.max_iterations)
                else:
                    z = cg(operator.H, grad_output[0], tolerance=tol_, max_iterations=ctx.max_iterations)
            if any(zi.isnan().any() for zi in z):
                raise RuntimeError('NaN in ConjugateGradientFunction.backward')
            with torch.enable_grad():
                residual = tuple(r - ax for r, ax in zip(rhs, operator(*(s.detach() for s in solution)), strict=True))
            grads = torch.autograd.grad(outputs=residual, inputs=inputs_with_grad, grad_outputs=z, allow_unused=True)
            grad_iter = iter(grads)
        else:
            grad_iter = iter(())

        grad_input = tuple(next(grad_iter) if need else None for need in ctx.needs_input_grad[2:])
        return (None, None, *grad_input)  # operator_factory, rhs_factory, *inputs


class ConjugateGradientOp(torch.nn.Module):
    r"""Solves a linear positive semidefinite system with the conjugate gradient method.

    Solves :math:`A x = b` where :math:`A` is a linear operator or a matrix of linear operators ,
    :math:`b` is a  tensor or a tuple of tensors.

    The operator is autograd differentiable using implicit differentiation. This is useful for including CG within a
    network [MODL]_, [PINQI]_.
    If this is not needed for your application, consider using `mrpro.algorithms.optimizers.cg` directly.

    References
    ----------
    .. [MODL] Aggarwal, H. K., et al. MoDL: Model-based deep learning architecture for inverse problems.
       (2018) IEEE TMI 2018, 38(2), 394-405. https://arxiv.org/abs/1712.02862
    .. [PINQI] Zimmermann, F. F., Kolbitsch, C., Schuenke, P., & Kofler, A. PINQI: an end-to-end physics-informed
       approach to learned quantitative MRI reconstruction. IEEE TCI 2024,  https://arxiv.org/abs/2306.11023
    """

    def __init__(
        self,
        operator_factory: Callable[..., LinearOperatorMatrix | LinearOperator],
        rhs_factory: Callable[..., tuple[torch.Tensor, ...]],
        implicit_backward: bool = True,
        tolerance: float = 1e-6,
        max_iterations: int = 100,
    ):
        r"""Initialize a conjugate gradient operator.

        Both the operator and the right-hand side are given as factory functions.
        The arguments given to the operator when calling it are passed to the factory functions.

        .. warning::
            If implicit_backward is `True`, `tolerance` and `max_iterations` should be chosen such that the cg algorithm
            converges, otherwise the backward will be wrong.

        **Example: Regularized Least Squares**

        Consider the regularized least squares problem:
        :math:`\min_x \|A x - y\|_2^2 + \alpha \|x - x_0\|_2^2`.

        The normal equations are :math:`(A^H A + \alpha I) x = A^H y + \alpha x_0`.
        This can be solved using the ConjugateGradientOp as follows:

        .. code-block:: python

            operator_factory = lambda alpha, x0, b: A.gram + alpha
            rhs_factory = lambda alpha, x0, b: A.H(b)[0] + alpha * x0
            op = ConjugateGradientOp(operator_factory, rhs_factory)
            solution = op(alpha, x0, b)

        Parameters
        ----------
        operator_factory
            A factory function that returns the operator :math:`A`.
            Should return either a `LinearOperatorMatrix` or a `LinearOperator`.
        rhs_factory
            A factory function that returns the right-hand side :math:`b`
            Should return a tuple of tensors.
        implicit_backward
            If `True`, the backward pass is done using implicit differentiation.
            If `False`, the backward pass is done using unrolling the CG loop.
        tolerance
            The tolerance for the conjugate gradient method. The tolerance is relative
            to the norm of the right-hand side. The same relative tolerance is used in the
            backward pass if using implicit differentiation.
        max_iterations
            The maximum number of iterations for the conjugate gradient method.
            The same maximum number of iterations is used in the backward pass if using
            implicit differentiation.
        """
        super().__init__()
        self.operator_factory = operator_factory
        self.rhs_factory = rhs_factory
        self.implicit_backward = implicit_backward
        self.tolerance = tolerance
        self.max_iterations = max_iterations

    def __call__(
        self, *parameters: torch.Tensor, initial_value: tuple[torch.Tensor, ...] | None = None
    ) -> tuple[torch.Tensor, ...]:
        """Solve the linear system using the conjugate gradient method.

        Parameters
        ----------
        parameters
            The parameters passed to the operator and right-hand side factory functions.
        initial_value
            The initial value for the conjugate gradient method.
            If `None`, the initial value is set to zero.

        Returns
        -------
            The solution of the linear system.
        """
        return super().__call__(*parameters, initial_value=initial_value)

    def forward(
        self,
        *parameters: torch.Tensor,
        initial_value: tuple[torch.Tensor, ...] | None = None,
    ) -> tuple[torch.Tensor, ...]:
        """Solve the linear system using the conjugate gradient method.

        .. note::
            Prefer calling the instance of the ConjugateGradientOp as ``operator(x)`` over directly calling this method.
            See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        if self.implicit_backward:
            solution = ConjugateGradientFunction.apply(self.operator_factory, self.rhs_factory, *parameters)
        else:  # unrolled CG
            op = self.operator_factory(*parameters)
            rhs = self.rhs_factory(*parameters)
            rhs_norm = sum((r.abs().square().sum() for r in rhs), torch.tensor(0.0)).sqrt().item()
            tol_ = self.tolerance * rhs_norm
            if isinstance(op, LinearOperator):
                if len(rhs) != 1:
                    raise ValueError('LinearOperator requires a single right-hand side tensor.')
                if initial_value is not None and len(initial_value) != 1:
                    raise ValueError('LinearOperator requires a single initial value tensor.')
                solution = cg(op, rhs, initial_value=initial_value, tolerance=tol_, max_iterations=self.max_iterations)
            else:
                solution = cg(op, rhs, initial_value=initial_value, tolerance=tol_, max_iterations=self.max_iterations)
        return solution
